# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import, print_function
import os
import sys
import signal

import argparse
from .. import __version__ as chemfp_version
from .. import ChemFPError, ParseError
from .. import io

try:
    import chemfp_converters
except ImportError:
    has_chemfp_converters = False
else:
    has_chemfp_converters = True

######
# Code extracted from Python's argparse.py. Distributed under the Python license.
# Before Python 3.4, argparse --version sent the message to stderr, not stdout.
# The bug was fixed in https://bugs.python.org/issue18920 .
# chemfp supports Python 2.7 and Python 3.5+.
# This is a back-port to version 2.7

version_action = "version"
if sys.version_info.major == 2:
    class ChemfpVersionAction(argparse._VersionAction):
        def __call__(self, parser, namespace, values, option_string=None):
            version = self.version
            if version is None:
                version = parser.version
            formatter = parser._get_formatter()
            formatter.add_text(version)
            parser._print_message(formatter.format_help(), sys.stdout)
            parser.exit()

    version_action = ChemfpVersionAction
    
def add_version(parser):
    parser.add_argument("--version", action=version_action,
                        version="%(prog)s " + chemfp_version)
    

######

class CallFunctionAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(CallFunctionAction, self).__init__(option_strings, dest, nargs=0, **kwargs)
    def __call__(self, parser, namespace, values, option_string=None):
        self.default()
        raise SystemExit(0)

######

def license_check():
    from .. import is_licensed, get_license_date
    if is_licensed():
        year, month, day = get_license_date()
        if year > 3000:
            # I'll let the future worry about this warning.
            sys.stdout.write("No license key required for this version of chemfp.\n")
        else:
            # Give a warning about possible problems
            sys.stdout.write("The license is valid until %4d-%02d-%02d.\n" % (
                year, month, day))
            check_license_date(sys.stdout)
        raise SystemExit(0)
            
    sys.stdout.write("Unable to use all chemfp functionality.\n")
    check_license_failure(None, sys.stdout)
    raise SystemExit(1)
    

def add_license_check(parser):
    parser.add_argument("--license-check", action=CallFunctionAction,
                            default=license_check,
                            help="Check the license and report results to stdout.")
                            
######

def mutual_exclusion(parser, args, default, groups):
    true_groups = []
    for g in groups:
        x = getattr(args, g)
        if x == "":
            # --type and --using take string arguments but the
            # empty string shouldn't be accepted as False
            true_groups.append(g)
        elif getattr(args, g):
            true_groups.append(g)

    if not true_groups:
        setattr(args, default, True)
    elif len(true_groups) == 1:
        pass
    else:
        parser.error("Cannot specify both --%s and --%s" % (true_groups[0], true_groups[1]))


class AbortHandler(object):
    def error(self, msg, location):
        raise ParseError("Each of the first 100 records contained errors. Final error: %s" % (msg,), location)

def cleanup_ids(reader, id_tag, error_handler, location):
    try:
        err_count = 0
        for recno, (id, fp) in enumerate(reader):
            if not id:
                if recno == err_count == 99:
                    error_handler = AbortHandler()
                err_count += 1

                if location.record_format == "sdf":
                    if id_tag is None:
                        error_handler.error("Missing title in SD record", location)
                    else:
                        error_handler.error("Missing id tag %r in SD record" % (id_tag,), location)
                elif location.record_format == "smi":
                    error_handler.error("Missing SMILES identifier (second column)", location)
                else:
                    error_handler.error("Missing identifier", location)
                continue

            else:
                # This assumes that these characters are rare
                if "\n" in id:
                    id = id.partition("\n")[0]  # get up to the newline
                if "\r" in id:
                    id = id.replace("\r", "")
                if "\t" in id:
                    id = id.replace("\t", "")
                if "\0" in id:
                    id = id.replace("\0", "")
                if " " in id:
                    id = id.strip()

                if not id:
                    if recno == err_count == 99:
                        error_handler = AbortHandler()
                    err_count += 1
                    if location.record_format == "sdf":
                        if id_tag is None:
                            error_handler.error("Empty title in SD record after cleanup", location)
                        else:
                            error_handler.error("Empty id tag %r in SD record after cleanup"
                                                % (id_tag,), location)
                    elif location.record_format == "smi":
                        error_handler.error("Empty SMILES identifier (second column) after cleanup", location)
                    else:
                        error_handler.error("Empty identifier after cleanup", location)
                    continue
                else:
                    yield id, fp
    finally:
        reader.close()
    

def sys_exit_opener(fptype, source, format, id_tag, reader_settings, error_handler, location):
    # Normalize to the format type associated with this source/format combination.
    format = fptype.toolkit.get_input_format_from_source(source, format)
    try:
        reader_args = format.get_reader_args_from_text_settings(reader_settings)
    except ValueError as err:
        sys.stderr.write("Could not process a -R option: %s. Exiting.\n" % err)
        raise SystemExit(1)
        
    try:
        return fptype.read_molecule_fingerprints(source, format, id_tag, reader_args=reader_args,
                                                 errors=error_handler, location=location)
    except (IOError, ChemFPError, ValueError) as err:
        sys.stderr.write("Problem reading structure fingerprints: %s. Exiting.\n" % err)
        raise SystemExit(1)

## def iter_all_sources(fptype, filenames, format, id_tag, errors):
##     for filename in filenames:
##         reader = sys_exit_opener(fptype, filename, format, id_tag, errors)
##         for id, fp in reader:
##             yield id, fp

def read_multifile_structure_fingerprints(fptype, filenames, format, id_tag, reader_settings, errors):
    error_handler = io.get_parse_error_handler(errors)
    if errors == "strict":
        id_error_handler = error_handler
    else:
        id_error_handler = io.get_parse_error_handler("report")
    
    if not filenames:
        location = io.Location.from_source(None)
        reader = sys_exit_opener(fptype, None, format, id_tag, reader_settings, error_handler, location)
        cleanup_reader = cleanup_ids(reader, id_tag, id_error_handler, location)
        return reader.metadata, cleanup_reader

    if len(filenames) == 1:
        location = io.Location.from_source(filenames[0])
        reader = sys_exit_opener(fptype, filenames[0], format, id_tag, reader_settings, error_handler, location)
        cleanup_reader = cleanup_ids(reader, id_tag, id_error_handler, location)
        return reader.metadata, cleanup_reader

    location = io.Location.from_source(filenames[0])
    reader = sys_exit_opener(fptype, filenames[0], format, id_tag, reader_settings, error_handler, location)
    reader.metadata.sources = filenames

    def multifile_reader(reader):
        # This is very annoying. The gzip reader fails if the file
        # isn't in the correct format but the check doesn't occur
        # until someone tries to read a byte. The error message is the
        # un-helpful generic "Not a gzipped file", without the
        # filename. I want to report that error, but do so carefully.

        cleanup_reader = cleanup_ids(reader, id_tag, id_error_handler, location)
        try:
            try:
                # Handle the first filename
                filename = filenames[0]

                for id, fp in cleanup_reader:
                    yield id, fp

                # Handle the remaining filenames
                for filename in filenames[1:]:
                    location.filename = filename
                    reader = sys_exit_opener(fptype, filename, format, id_tag, reader_settings,
                                             error_handler, location)
                    cleanup_reader = cleanup_ids(reader, id_tag, id_error_handler, location)
                    for id, fp in cleanup_reader:
                        yield id, fp

            except IOError as err:
                if err.filename is None:
                    raise IOError("%s: %r" % (err, filename))
                raise
            except EOFError as err:
                raise EOFError("%s: %r" % (err, filename))
        finally:
            # We are passed a reader which we must close.
            cleanup_reader.close()
            

    return reader.metadata, multifile_reader(reader)

def is_valid_tag(tag):
    if tag is None:
        return True
    for c in "<>\r\n":
        if c in tag:
            return False
    return True


def get_reader_settings(parser, args):
    if not args.reader_settings:
        return {}
    
    reader_settings = {}
    for r_option in args.reader_settings:
        left, mid, right = r_option.partition("=")
        if mid != "=":
            parser.error("-R setting %r must be in the format NAME=VALUE" % (r_option,))
        reader_settings[left] = right
    return reader_settings

def get_writer_settings(parser, args):
    if not args.writer_settings:
        return {}
    
    writer_settings = {}
    for r_option in args.writer_settings:
        left, mid, right = r_option.partition("=")
        if mid != "=":
            parser.error("-R setting %r must be in the format NAME=VALUE" % (r_option,))
        writer_settings[left] = right
    return writer_settings

def check_filenames(parser, toolkit, format, filenames):
    if not filenames:
        # Read from stdin.
        return None
    for filename in filenames:
        if not os.path.exists(filename):
            parser.error("Structure file %r does not exist" % (filename,))
        try:
            toolkit.get_input_format_from_source(source=filename, format=format)
        except ValueError as err:
            parser.error(str(err))

def add_argument_to_argparse(argument, parser):
    """Add information about the given parameter to an argparse parser

    This calls parser.add_argument() using "--$NAME" as the argument name
    and along with values for `type`, `default`, `action`, `metavar`,
    and `help`.
    """
    # Wrapper function to convert the decoder() API into
    # the argparse API.
    def parse_argument(s):
        try:
            return argument.decoder(s)
        except ValueError as err:
            raise argparse.ArgumentError(None, "--%s %s" % (argument.name, err))

    argparse_kwargs = dict(type=parse_argument,
                           default=argument.default,
                           action=argument.action,
                           metavar=argument.metavar,
                           help=argument.help)

    parser.add_argument("--" + argument.name, **argparse_kwargs)


class DelimiterAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.reader_settings is None:
            namespace.reader_settings = []
        namespace.reader_settings.append("delimiter=%s" % (values,))
        
class HasHeaderAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.reader_settings is None:
            namespace.reader_settings = []
        namespace.reader_settings.append("has_header=1")


def add_delimiter_alias(parser):
    parser.add_argument("--delimiter", choices=("tab", "whitespace", "to-eol", "space"),
                        action=DelimiterAction,
                        help="delimiter style for SMILES and InChI files. Alias for '-R delimiter=VALUE'.")
def add_has_header_alias(parser):
    parser.add_argument("--has-header", action=HasHeaderAction, nargs=0,
                        help="Skip the first line of a SMILES or InChI file Alias for '-R has_header=1'")
    
def add_reader_aliases(parser):
    add_delimiter_alias(parser)
    add_has_header_alias(parser)


def add_type_and_using(parser):
    parser.add_argument("--type", metavar="TYPE_STRING",
                            help="Specify a chemfp type string")
    parser.add_argument("--using", metavar="FILENAME",
                            help="Get the fingerprint type from the metadata of a fingerprint file")

def get_fptype_from_type_arg(parser, typestr, progname, toolkit_name):
    from .. import get_fingerprint_type
    try:
        fptype = get_fingerprint_type(typestr)
    except ValueError as err:
        sys.stderr.write("Unable to use the specified --type: %s\n" % (err,))
        raise SystemExit(1)
    
    if fptype.toolkit is None:
        sys.stderr.write("Unable to use the specified --type: %r is not associated with a toolkit\n" % (
            typestr,))
        raise SystemExit(1)
    
    if fptype.toolkit.name != toolkit_name:
        sys.stderr.write(
            "WARNING: %s uses the %s toolkit but --type %r expects fingerprints using the %s toolkit.\n" % (
                progname, toolkit_name, typestr, fptype.toolkit.name))
    return fptype

def get_fptype_from_using_arg(parser, using_str, progname, toolkit_name):
    from .. import open as chemfp_open, get_fingerprint_type, ChemFPError
    try:
        reader = chemfp_open(using_str)
    except ChemFPError as err:
        sys.stderr.write("Cannot open --using file %r: %s\n" % (using_str, err))
        raise SystemExit(1)
    except (OSError, IOError, ValueError) as err:
        sys.stderr.write("Cannot open --using file %r: %s\n" % (using_str, err))
        raise SystemExit(1)

    with reader:
        typestr = reader.metadata.type
        if typestr is None:
            sys.stderr.write("No fingerprint type found in the --using file %r\n" % (
                using_str,))
            raise SystemExit(1)
        try:
            fptype = get_fingerprint_type(typestr)
        except ValueError as err:
            sys.stderr.write("Unable to use the fingerprint type from the --using file %r: %s\n" % (
                using_str, err))
            raise SystemExit(1)
            
    if fptype.toolkit is None:
        sys.stderr.write("Unable to use the fingerprint type in --using file %r: %r is not associated with a toolkit\n" % (
            using_str, typestr))
        raise SystemExit(1)
    
    if fptype.toolkit.name != toolkit_name:
        sys.stderr.write(
            "WARNING: Fingerprint type %r from the --using file %r is based on the %s toolkit.\n" % (
                typestr, using_str, fptype.toolkit.name))
        sys.stderr.write(
            "WARNING: %s expects fingerprints based on the %s toolkit.\n" % (
                progname, toolkit_name))
    return fptype

def run(main_func):
    signal.signal(signal.SIGINT, signal.SIG_DFL)  # Allow ^C to kill the process.
    signal.signal(signal.SIGPIPE, signal.SIG_DFL) # Allow the output pipe to be closed
    try:
        main_func()
    except KeyboardInterrupt:
        raise SystemExit(2)

def die(msg):
    sys.stderr.write(msg)
    sys.stderr.write("\n")
    sys.stderr.flush()
    raise SystemExit(1)


def writer_failure(tool_name, msg):
    check_license_failure(msg)
    sys.stderr.write("Exiting %s.\n" % (tool_name,))
    raise SystemExit(1)
    

def check_license_failure(msg, outfile=sys.stderr):
    from .. import get_license_date, __version__
    
    def get_reason():
        date = get_license_date()
        if date is None:
            # It's also possible someone could have embedded an incorrect
            # license key in the code. That shouldn't happen. In any case,
            # $CHEMFP_LICENSE will override that case.
            if not "CHEMFP_LICENSE" in os.environ:
                return "The CHEMFP_LICENSE environment variable is not defined."
            license = os.environ["CHEMFP_LICENSE"]
            license_date = license[:8]
            if len(license) < 40 or not license_date.isdigit() or "@" not in license:
                return "$CHEMFP_LICENSE does not contain a chemfp license."
            return "$CHEMFP_LICENSE does not not appear to be valid."
        year, month, day = date
        if "CHEMFP_LICENSE" in os.environ:
            where = "The license in $CHEMFP_LICENSE"
        else:
            where = "The embedded license"
        return "%s expired on %04d-%02d-%02d." % (where, year, month, day)

    errmsg = get_reason()
    if msg is not None:
        outfile.write(msg)
        if not msg.endswith("\n"):
            outfile.write("\n")
        
    outfile.write("The license key check failed. %s\n" % (errmsg,))
    outfile.write("Email sales@dalkescientific.com to purchase a license key or request a demo license key.\n")


def check_license_date(outfile=sys.stderr):
    from .. import get_license_date
    year, month, day = get_license_date()
    if year > 3000:
        # I'll let the future worry about this warning.
        return
    import datetime
    delta = datetime.datetime(year, month, day, 23, 59, 59) - datetime.datetime.utcnow()
    if delta.days < 7:
        if delta.days == 0:
            outfile.write("WARNING: Your chemfp license expires TODAY.\n")
        elif delta.days == 1:
            outfile.write("WARNING: Your chemfp license expires TOMORROW.\n")
        else:
            outfile.write("WARNING: Your chemfp license expires in %d days.\n" % (delta.days,))
    
