# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import with_statement, print_function

import sys
import argparse

from .. import ChemFPError, open_fingerprint_writer, has_fingerprint_family, get_fingerprint_family
from .. import rdkit_types, rdkit_toolkit
from .. import get_fingerprint_type, open as chemfp_open
from . import cmdsupport

########### Configure the command-line parser

epilog = ""

if not has_fingerprint_family("RDKit-Pattern"):
    epilog += """\
    
This version of RDKit does not support the 'Pattern' fingerprints.
"""

if not has_fingerprint_family("RDKit-Avalon"):
    epilog += """\
    
This version of RDKit does not support the Avalon fingerprints. To use them,
reinstall RDKit with pyAvalonTools.
"""
    

epilog += """\

This program guesses the input structure format and the compression
based on the filename extension. If the guess fails then it assumes
the input is an uncompressed SMILES file.

If the data comes from stdin, or the guess based on extension name is
wrong, then use "--in" to change the default input format.

Use the '-R' reader arguments option to pass in format-specific structure
reader arguments. The details depend on the specific format.

Use the command-line option `--help-formats` to display a list of
available formats and reader arguments.
"""


def help_formats():
    print("""\
These are the structure file formats that chemfp can read when using
the RDKit toolkit.

By default, chemfp uses the filename extension to determine the format
type. If the filename ends with ".gz" or ".zst" then it is intepreted
as a gzip or Zstandard compressed file, and the second-to-last
extension is used to determine the format type. Unknown or unsupported
extensions are interpreted as a SMILES file.
""")
    try:
        import zstandard
    except ImportError:
        print("""\
Note: To enable Zstandard compression, please install the "zstandard"
Python package from https://pypi.org/project/zstandard/ .
""")

    print("""\
You may instead specify the file format by name (see below), which is
especially important when reading from stdin, which has no associated
filename extension.

The supported filename extensions are:
""")
    print("File Type".center(15) + " Extension(s)")
    print("==========".center(15) + " =============")
    format_lines = []
    format_data = (
            ("SMILES", ["smi", "can", "usm"]),
            ("SDF", ["sdf"]),
            ("InChI", ["inchi"]),
            ("Tripos Mol2", ["mol2"]),
            ("PDB", ["pdb"]),
            ("Maestro", ["mae"]),
            ("FASTA", ["fasta"]),
            )
    for format_type, format_names in format_data:
        try:
            rdkit_toolkit.get_format(format_names[0])
        except ValueError:
            msg = (format_type.center(15) + " (Unsupported in this RDKit installation)")
            format_lines.append(
                format_type.center(15) + " " +
                ", ".join(format_names) + " (Unsupported in this RDKit installation)")
        else:
            extensions = set()
            for name in format_names:
                extensions.update(rdkit_toolkit.get_format(name).extensions)
            msg = format_type.center(15) + " " + ", ".join(sorted(extensions))
            format_lines.append(format_type.center(15) + " " + ", ".join(format_names))
        print(msg)

    print("""\

The format can also be specified by name using the '--in' option:

   File Type    Format name (append .gz or .zst if compressed)
   ==========   ==============================================""")
    for format_line in format_lines:
        print(format_line)
                      
    print("""\

The input format parsers can be configured with the "-R" option. For
example, the following reader arguments tell the SMILES readers that
the fields are whitespace delimited and the first line is a header.

   -R delimiter=whitespace -R has_header=true

All of the input formats implement the 'sanitize' option, which is
enabled by default. Use "-R sanitize=false" to disable sanitization.

The SMILES format parsers use two additional reader arguments:
   * 'delimiter' specifies the delimiter type. The default is 'to-eol'.
     The other values are 'tab', 'whitespace', 'space' and 'native'.
     Use "-R delimiter=native" to match RDKit's native delimiter
     style, which is 'whitespace'.
   * 'has_header', if false will skip the first line
     of the SMILES file (because it is a header line).

The SDF format parser supports two additional reader arguments:
   * 'strictParsing', if false will disable strict parsing
   * 'removeHs', if false will keep all of the hydrogens

The InChI format parser supports four additional reader arguments:
   * 'delimiter' works the same as it does for the SMILES formats
   * 'removeHs' works the same as it does for the SDF format
   * 'treatWarningAsError', if true treats all warnings as errors
   * 'logLevel' specifies the RDKit/InChI library log level, as an integer

The Tripos Mol2 format parser supports two additional reader arguments:
   * 'removeHs' works the same as it does for the SDF format
   * 'cleanupSubstructures' if false disables standardizing
      some substructures found in Mol2 files

The PDB format parser supports three additional reader arguments:
   * 'removeHs' works the same as it does for the SDF format
   * 'flavor', an input parameter with no documented meaning
   * 'proximityBonding', if false will disable automatic
       automatic proximity bonding

The Maestro format parser supports one additional reader argument:
   * 'removeHs' works the same as it does for the SDF format

The FASTA format parser supports one additional reader argument:
   * 'flavor', an integer from 0 to 9. The values mean:
       0 - the sequence contains L-amino acids 
       1 - allow lowercase for D-amino acids
       2 - RNA with no cap        6 - DNA with no cap
       3 - RNA with 5' cap        7 - DNA with 5' cap
       4 - RNA with 3' cap        8 - DNA with 3' cap
       5 - RNA with both caps     9 - DNA with both caps
""")

if 0:
    # Check that there are no duplicate argument registrations
    _add_argument = argparse._ActionsContainer.add_argument
    seen_args = set()
    def new_add_argument(self, *args, **kwargs):
        for arg in args:
            if isinstance(arg, str) and arg[:1] == "-":
                assert arg not in seen_args, ("duplicate", arg)
                
                seen_args.add(arg)
        return _add_argument(self, *args, **kwargs)
    argparse._ActionsContainer.add_argument = new_add_argument
    
parser = argparse.ArgumentParser(
    description="Generate FPS or FPB fingerprints from a structure file using RDKit",
    epilog=epilog,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    conflict_handler="resolve",
    )


parser.add_argument("--type", metavar="TYPE_STRING",
                       help="Specify a chemfp type string")
parser.add_argument("--using", metavar="FILENAME",
                       help="Get the fingerprint type from the metadata of a fingerprint file")


all_arguments = rdkit_types.all_arguments

# --RDK and --morgan both have fpSize but argparse doesn't allow the
# same option in different groups. Especially with different defaults.

generic_group = parser.add_argument_group(
    "Common Parameters (used by more than one fingerprint type)"
    )

cmdsupport.add_argument_to_argparse(rdkit_types.generic_fpSize, generic_group)
cmdsupport.add_argument_to_argparse(rdkit_types.generic_radius, generic_group)
cmdsupport.add_argument_to_argparse(rdkit_types.generic_nBitsPerEntry, generic_group)
cmdsupport.add_argument_to_argparse(rdkit_types.generic_includeChirality, generic_group)


def fromAtom_list(s):
    t = s.strip()
    if not t:
        raise argparse.ArgumentTypeError(
            "must contain a comma-separated list of atom indices")

    terms = t.split(",")
    seen = set()
    for term in terms:
        if not term.isdigit():
            raise argparse.ArgumentTypeError(
                "term %r must be a non-negative integer"
                % (term,))
        seen.add(int(term))
    return sorted(seen)

##from_group = parser.add_mutually_exclusive_group()
## from_group.add_argument("--all-atoms", action="store_true")
## from_group.add_argument("--unique-atoms", action="store_true")
generic_group.add_argument(
    "--from-atoms", dest="fromAtoms", metavar="INT,INT,...",
    type=fromAtom_list, default=None,
    help="fingerprint generation must use these atom indices (out of range indices are ignored)")

rdk_group = parser.add_argument_group(
    "RDKit topological fingerprints",
    description =
    "Branched or linear hash fingerprint.\n"
    "Uses --fpSize and --fromAtoms plus:")
rdk_group.add_argument("--RDK", action="store_true",
                       help="generate RDK fingerprints (default)")
rdkit_fingerprint_arguments = rdkit_types.rdkit_fingerprint_arguments
cmdsupport.add_argument_to_argparse(rdkit_fingerprint_arguments["minPath"], rdk_group)
cmdsupport.add_argument_to_argparse(rdkit_fingerprint_arguments["maxPath"], rdk_group)
cmdsupport.add_argument_to_argparse(rdkit_fingerprint_arguments["nBitsPerHash"], rdk_group)
cmdsupport.add_argument_to_argparse(rdkit_fingerprint_arguments["useHs"], rdk_group)
cmdsupport.add_argument_to_argparse(rdkit_fingerprint_arguments["branchedPaths"], rdk_group)
cmdsupport.add_argument_to_argparse(rdkit_fingerprint_arguments["useBondOrder"], rdk_group)


morgan_group = parser.add_argument_group(
    "RDKit Morgan fingerprints",
    description =
    "Circular fingerprints similar to ECFP or FCFP fingerprints.\n"
    "Uses --fpSize, --radius, and --fromAtoms plus:")
morgan_group.add_argument("--morgan", action="store_true",
                          help="generate Morgan fingerprints")

#cmdsupport.add_argument_to_argparse(all_arguments["radius"], morgan_group)
morgan_arguments = rdkit_types.morgan_arguments
cmdsupport.add_argument_to_argparse(morgan_arguments["useFeatures"], morgan_group)
cmdsupport.add_argument_to_argparse(morgan_arguments["useChirality"], morgan_group)
cmdsupport.add_argument_to_argparse(morgan_arguments["useBondTypes"], morgan_group)
cmdsupport.add_argument_to_argparse(morgan_arguments["includeRedundantEnvironments"], morgan_group)

torsion_group = parser.add_argument_group(
    "RDKit Topological Torsion fingerprints",
    description=
    "See Nilakantan et al., JCICS 27, 82-85 (1987).\n"
    "Uses --fpSize, --nBitsPerEntry, --includeChirality, and --fromAtoms plus:")
torsion_fingerprint_parameters = rdkit_types.torsion_fingerprint_parameters
torsion_group.add_argument("--torsions", action="store_true",
                           help="generate Topological Torsion fingerprints")
cmdsupport.add_argument_to_argparse(torsion_fingerprint_parameters["targetSize"], torsion_group)
#cmdsupport.add_argument_to_argparse(torsion_fingerprint_parameters["nBitsPerEntry"], torsion_group)
#cmdsupport.add_argument_to_argparse(torsion_fingerprint_parameters["includeChirality"], torsion_group)

pair_group = parser.add_argument_group(
    "RDKit Atom Pair fingerprints",
    description =
    "See Carhart et al., JCICS 25, 64-73 (1985).\n"
    "Uses --fpSize, --nBitsPerEntry, --includeChirality, and --fromAtoms plus:")
pair_group.add_argument("--pairs", action="store_true",
                        help="generate Atom Pair fingerprints")
cmdsupport.add_argument_to_argparse(rdkit_types.atom_pair_arguments["minLength"], pair_group)
cmdsupport.add_argument_to_argparse(rdkit_types.atom_pair_arguments["maxLength"], pair_group)
#cmdsupport.add_argument_to_argparse(rdkit_types.atom_pair_arguments["nBitsPerEntry"], pair_group)
#cmdsupport.add_argument_to_argparse(rdkit_types.atom_pair_arguments["includeChirality"], pair_group)
cmdsupport.add_argument_to_argparse(rdkit_types.atom_pair_arguments["use2D"], pair_group)



maccs_group = parser.add_argument_group("166 bit MACCS substructure keys")
maccs_group.add_argument(
    "--maccs166", action="store_true", help="generate MACCS fingerprints")


avalon_help = "generate Avalon fingerprints"
if has_fingerprint_family("RDKit-Avalon"):
    has_avalon = True
else:
    has_avalon = False
    avalon_help += " (not available for this RDKit install)"
avalon_arguments = rdkit_types.avalon_arguments
avalon_group = parser.add_argument_group(
    "Avalon fingerprints",
    description =
    "Fingerprints from the Avalon toolkit.\n"
    "Uses --fpSize plus:")
avalon_group.add_argument(
    "--avalon", action="store_true", help=avalon_help)
cmdsupport.add_argument_to_argparse(avalon_arguments["isQuery"], avalon_group)
cmdsupport.add_argument_to_argparse(avalon_arguments["bitFlags"], avalon_group)
    
secfp_help = "generate SECFP fingerprints"
if has_fingerprint_family("RDKit-SECFP"):
    has_secfp = True
else:
    has_secfp = False
    secfp_help += " (not available for this RDKit install)"
secfp_arguments = rdkit_types.secfp_arguments
secfp_group = parser.add_argument_group(
    "SECFP fingerprints",
    description =
    "A circular fingerprint based on fragment SMILES instead of hashing.\n"
    "Uses --fpSize and --radius plus:")
secfp_group.add_argument(
    "--secfp", action="store_true", help=secfp_help)
cmdsupport.add_argument_to_argparse(secfp_arguments["rings"], secfp_group)
cmdsupport.add_argument_to_argparse(secfp_arguments["isomeric"], secfp_group)
cmdsupport.add_argument_to_argparse(secfp_arguments["kekulize"], secfp_group)
cmdsupport.add_argument_to_argparse(secfp_arguments["min_radius"], secfp_group)

# Use "--min-radius" as an alias for "--min_radius"
# Why? I found myself making the mistake.
# If only the SECFP API used "minRadius" instead of "min_radius". Oh well.
_secfp_min_radius = secfp_arguments["min_radius"]
secfp_group.add_argument("--min-radius", help=argparse.SUPPRESS,
                             dest="min_radius",
                             type=_secfp_min_radius.decoder)

pattern_group = parser.add_argument_group(
    "RDKit Pattern fingerprints",
    description="Fingerprints for substructure search screening.")
if has_fingerprint_family("RDKit-Pattern"):
    pattern_group.add_argument("--pattern", action="store_true",
                               help="generate (substructure) pattern fingerprints")
else:
    pattern_group.add_argument("--pattern", action="store_true",
                               help="(substructure patterns are not available in this version of RDKit)")


substruct_group = parser.add_argument_group(
    "chemfp's version of the 881 bit PubChem substructure keys")
substruct_group.add_argument(
    "--substruct", action="store_true", help="generate ChemFP substructure fingerprints")

rdmaccs_group = parser.add_argument_group(
    "chemfp's version of the 166 bit RDKit/MACCS keys")
rdmaccs_group.add_argument(
    "--rdmaccs", "--rdmaccs/2", action="store_true", help="generate 166 bit RDKit/MACCS fingerprints (version 2)")
rdmaccs_group.add_argument(
    "--rdmaccs/1", dest="rdmaccs/1", action="store_true", help="use the version 1 definition for --rdmaccs")

parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag name containing the record id (SD files only)")

cmdsupport.add_type_and_using(parser)

parser.add_argument(
    "--in", metavar="FORMAT", dest="format",
    help="input structure format (default guesses from filename)")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")
parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "fps.zst", "fpb", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")


parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")

parser.add_argument(
    "--help-formats", action=cmdsupport.CallFunctionAction,
    default=help_formats,
    help="list the available formats and reader arguments")
parser.add_argument(
    "-R", metavar="NAME=VALUE", dest="reader_settings", action="append",
    help="specify a reader argument")
cmdsupport.add_reader_aliases(parser)

cmdsupport.add_version(parser)

parser.add_argument(
    "filenames", nargs="*", help="input structure files (default is stdin)")

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)

    cmdsupport.mutual_exclusion(parser, args, "RDK",
                                ("maccs166", "RDK", "substruct", "rdmaccs", "rdmaccs/1", "morgan",
                                 "torsions", "pairs", "pattern", "avalon", "secfp",
                                  "type", "using"))

    if args.maccs166:
        fptype = get_fingerprint_family("RDKit-MACCS166")()
    elif args.RDK:
        fpSize = args.fpSize or rdkit_types.NUM_BITS
        minPath = args.minPath
        maxPath = args.maxPath
        nBitsPerHash = args.nBitsPerHash
        if maxPath < minPath:
            parser.error("--minPath must not be greater than --maxPath")

        useHs = args.useHs

        fptype = get_fingerprint_family("RDKit-Fingerprint")(
            minPath=minPath,
            maxPath=maxPath,
            fpSize=fpSize,
            nBitsPerHash=nBitsPerHash,
            useHs=useHs,
            branchedPaths=args.branchedPaths,
            useBondOrder=args.useBondOrder,

            fromAtoms=args.fromAtoms,
            )

    elif args.substruct:
        fptype = get_fingerprint_family("ChemFP-Substruct-RDKit")()
    elif args.rdmaccs:        
        fptype = get_fingerprint_family("RDMACCS-RDKit")()
    elif getattr(args, "rdmaccs/1"):
        fptype = get_fingerprint_family("RDMACCS-RDKit/1")()
    elif args.morgan:
        fptype = get_fingerprint_family("RDKit-Morgan")(
            radius=args.radius,
            fpSize=args.fpSize,
            useFeatures=args.useFeatures,
            useChirality=args.useChirality,
            useBondTypes=args.useBondTypes,
            fromAtoms=args.fromAtoms,
            includeRedundantEnvironments=args.includeRedundantEnvironments,
            )
    elif args.torsions:
        fptype = get_fingerprint_family("RDKit-Torsion")(
            fpSize=args.fpSize,
            targetSize=args.targetSize,
            fromAtoms=args.fromAtoms,
            nBitsPerEntry=args.nBitsPerEntry,
            includeChirality=args.includeChirality,
            )
    elif args.pairs:
        minLength = args.minLength
        maxLength = args.maxLength
        if maxLength < minLength:
            parser.error("--minLength must not be greater than --maxLength")
        fptype = get_fingerprint_family("RDKit-AtomPair")(
            fpSize=args.fpSize,
            minLength=minLength,
            maxLength=maxLength,
            fromAtoms=args.fromAtoms,
            nBitsPerEntry=args.nBitsPerEntry,
            includeChirality=args.includeChirality,
            use2D=args.use2D,
            )
    elif args.pattern:
        try:
            family = get_fingerprint_family("RDKit-Pattern")
        except ValueError:
            parser.error("--pattern fingerprints are not available in this version of RDKit")
        fptype = family(fpSize=args.fpSize)
    elif args.avalon:
        if has_avalon:
            fptype = get_fingerprint_family("RDKit-Avalon")(
                fpSize=args.fpSize,
                isQuery=args.isQuery,
                bitFlags=args.bitFlags)

        else:
            parser.error("Cannot generate --avalon fingerprints. The RDKit toolkit was not compiled with Avalon support.")
    elif args.secfp:
        if has_secfp:
            # Fill in the radius/min_radius so I can check if there will be a problem
            if args.radius is None:
                args.radius = rdkit_types.secfp_arguments["radius"].default
            if args.min_radius is None:
                args.min_radius = rdkit_types.secfp_arguments["min_radius"].default
            if args.radius < args.min_radius:
                parser.error("--min_radius must not be greater than --radius")
            fptype = get_fingerprint_family("RDKit-SECFP")(
                fpSize=args.fpSize,
                radius=args.radius,
                rings=args.rings,
                isomeric=args.isomeric,
                kekulize=args.kekulize,
                min_radius=args.min_radius)
        else:
            parser.error("Cannot generate --secfp fingerprints. The RDKit toolkit was not compiled with SECFP support.")
    elif args.type is not None:
        fptype = cmdsupport.get_fptype_from_type_arg(parser, args.type, "rdkit2fps", "rdkit")
    elif args.using is not None:
        fptype = cmdsupport.get_fptype_from_using_arg(parser, args.using, "rdkit2fps", "rdkit")
    else:
        raise AssertionError("Unknown fingerprinter")

    # Check that this is a supported format.
    try:
        rdkit_toolkit.get_input_format(args.format)
    except ValueError:
        parser.error("Unsupported format specifier: %r" % (args.format,))

    if not cmdsupport.is_valid_tag(args.id_tag):
        parser.error("Invalid id tag: %r" % (args.id_tag,))

    reader_settings = cmdsupport.get_reader_settings(parser, args)
    if "includeTags" not in reader_settings:
        # rdkit2fps doesn't need the SD tags to generate fingerprints
        reader_settings["includeTags"] = "False"

    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")
    
    cmdsupport.check_filenames(parser, rdkit_toolkit, args.format, args.filenames)

    # Ready the input reader/iterator
    metadata, reader = cmdsupport.read_multifile_structure_fingerprints(
        fptype, args.filenames, format = args.format, id_tag = args.id_tag, 
        reader_settings = reader_settings, errors = args.errors)

    try:
        try:
            writer = open_fingerprint_writer(args.output, metadata, format=args.output_format)
        except NotImplementedError as err:
            cmdsupport.writer_failure("rdkit2fps", str(err))
            
        except (IOError, ValueError) as err:
            msg = "Cannot open output fingerprint file: %s" % (err,)
            if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
                msg += ".\nUse --out to specify 'fps', 'fps.gz', 'fps.zst', 'fpb', or 'flush'."
            parser.error(msg)

        try:
            with writer:
                writer.write_fingerprints(reader)
        except (ChemFPError, IOError, EOFError) as err:
            sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
            raise SystemExit(1)
    finally:
        reader.close()

if __name__ == "__main__":
    main()

