# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import, print_function

import _chemfp
import sys
import itertools
import array

from . import Metadata, ParseError, FingerprintReader, FingerprintWriter
from . import fps_search
from . import io

from ._compat import next, raise_tb

BLOCKSIZE=100000

## class FPSParseError(ParseError):
##     def __init__(self, errcode, lineno, filename):
##         super(FPSParseError, self).__init__(errcode, lineno, filename)
##         self.errcode = errcode
##         self.lineno = lineno
##         self.filename = filename
##     def __repr__(self):
##         return "FPSParseError(%d, %d, %s)" % (self.errcode, self.lineno, self.filename)
##     def __str__(self):
##         msg = _chemfp.strerror(self.errcode)
##         msg += " at line %d" % (self.lineno,)
##         if self.filename is not None:
##             msg += " of %r" % (self.filename,)
##         return msg


def open_fps(source, format_name, compression, location=None):
    if format_name != "fps":
        raise ValueError("Unknown format %r" % (format_name,))

    infile, close = io.open_compressed_input(source, compression, "chemfp")
    if location is None:
        location = io.Location.from_source(source)
    try:
        #filename = io.get_filename(source)
        metadata, block, block_reader = read_header(infile, location)
        if block is None:
            block_reader = iter([])
        else:
            block_reader = itertools.chain([block], block_reader)
        return FPSReader(infile, close, metadata, location, block_reader)
    except Exception:
        infile.close() # Prevent a ResourceWarning if the file isn't in the right format
        raise


# This never buffers
def _read_blocks(infile):
    # ZstdDecompressionReader doesn't implement readline()
    # but there's no easy way to tell if I'm using it.
    has_readline = hasattr(infile, "has_readline")
    if has_readline:
        x = getattr(infile, "__class__", None)
        if x is not None:
            x = getattr(x, "__name__", None)
            if x == "ZstdDecompressionReader" and str(infile).startswith("<zstd."):
                has_readline = False
                
    if has_readline:
        # This seems to be slightly faster for uncompressed files
        return _read_blocks_with_readline(infile)
    else:
        # Need a workaround. Which can use 2-3x more memory.
        # But that shouldn't be a problem.
        return _read_blocks_without_readline(infile)

def _read_blocks_with_readline(infile):
    while 1:
        block = infile.read(BLOCKSIZE)
        if not block:
            break
        if block[-1:] == b"\n":
            yield block
            continue
        line = infile.readline()
        if not line:
            # Note: this might not end with a newline!
            yield block
            break
        yield block + line

def _read_blocks_without_readline(infile):
    block = infile.read(BLOCKSIZE)
    if not block:
        return
    
    while 1:
        if block[-1:] == b"\n":
            yield block
            block = infile.read(BLOCKSIZE)
            if not block:
                return
        else:
            while 1:
                new_block = infile.read(BLOCKSIZE)
                if not new_block:
                    yield block
                    return

                i = new_block.find(b"\n")
                if i == -1:
                    block += new_block
                else:
                    i += 1
                    yield block + new_block[:i]
                    block = new_block[i:]
                    break

class FPSReader(FingerprintReader):
    """FPS file reader

    This class implements the :class:`chemfp.FingerprintReader` API. It
    is also its own a context manager, which automatically closes the
    file when the manager exists.

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.Metadata` instance with information about the fingerprint type
       
    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance with parser location and state information
       
    .. py:attribute:: closed

       True if the file is open, else False
    
    The FPSReader.location only tracks the "lineno" variable.
    """
    _search = fps_search
    def __init__(self, infile, close, metadata, location, block_reader):
        self._infile = infile
        self._close = close
        self._filename = getattr(infile, "name", "<unknown>")
        self.metadata = metadata
        self.location = location
        if metadata.num_bytes is None:
            self._expected_hex_len = 0
        else:
            self._expected_hex_len = 2*metadata.num_bytes
        self._hex_len_source = "size in header"

        self._at_start = True
        self._it = None
        self._block_reader = block_reader
        self._iter = None

        self._record_iter = None

        self.closed = False

    def close(self):
        """Close the file"""
        if self._close is not None:
            self._close()
        self.closed = True

    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()
    
    def iter_blocks(self):
        """This is not part of the public API"""
        self._at_start = False
        return self._block_reader

    def _iter_blocks(self):
        if not self._at_start:
            raise TypeError("Already iterating")
        
        self._at_start = False

        if self._first_fp_block is None:
            return
        
        block_stream = _read_blocks(self._infile)
        yield self._first_fp_block
        for block in block_stream:
            yield block

    def iter_rows(self):
        """This is not part of the public API"""
        expected_hex_len = self._expected_hex_len
        
        location = self.location
        lineno = location.lineno
        def get_lineno():
            return lineno
        location.register(get_lineno=get_lineno)
            
        try:
            for block in self.iter_blocks():
                for line in block.splitlines(True):
                    lineno += 1
                    err = _chemfp.fps_line_validate(expected_hex_len, line)
                    if err:
                        raise ParseError(_chemfp.strerror(err), location)
                    yield line[:-1].split(b"\t")
        finally:
            location.save(lineno=lineno)

    def __next__(self):
        if self._record_iter is None:
            self._record_iter = self._make_record_iter()
        return next(self._record_iter)

    def next(self):
        "Return the next (id, fp) pair"
        if self._record_iter is None:
            self._record_iter = self._make_record_iter()
        return next(self._record_iter)
        

    def __iter__(self):
        "Iterate through the (id, fp) pairs"
        if self._record_iter is None:
            self._record_iter = iter(self._make_record_iter())
        return self._record_iter

    def _make_record_iter(self):
        expected_hex_len = self._expected_hex_len
        
        location = self.location
        lineno = location.lineno
        def get_lineno():
            return lineno
        location.register(get_lineno=get_lineno)

        try:
            for block in self.iter_blocks():
                for line in block.splitlines(True):
                    lineno += 1
                    err, id_fp = _chemfp.fps_parse_id_fp(expected_hex_len, line)
                    if err:
                        # Include the line?
                        raise ParseError(_chemfp.strerror(err), location)
                    yield id_fp
        finally:
            location.save(lineno=lineno)

    def _check_at_start(self):
        if not self._at_start:
            raise TypeError("FPS file is not at the start of the file; cannot search")


    def count_tanimoto_hits_fp(self, query_fp, threshold=0.7):
        """Count the fingerprints which are sufficiently similar to the query fingerprint

        Return the number of fingerprints in the reader which are
        at least *threshold* similar to the query fingerprint *query_fp*.

        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: integer count
        """
        self._check_at_start()
        return fps_search.count_tanimoto_hits_fp(query_fp, self, threshold)

    def count_tanimoto_hits_arena(self, queries, threshold=0.7):
        """Count the fingerprints which are sufficiently similar to each query fingerprint

        Returns a list containing a count for each query fingerprint
        in the *queries* arena. The count is the number of
        fingerprints in the reader which are at least *threshold*
        similar to the query fingerprint.

        The order of results is the same as the order of the queries.
        
        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: list of integer counts, one for each query
        """
        self._check_at_start()
        return fps_search.count_tanimoto_hits_arena(queries, self, threshold)

    def threshold_tanimoto_search_fp(self, query_fp, threshold=0.7):
        """Find the fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this reader which are at least
        *threshold* similar to the query fingerprint *query_fp*.  The
        hits are returned as a :class:`.SearchResult`, in arbitrary
        order.
        
        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        self._check_at_start()
        return fps_search.threshold_tanimoto_search_fp(query_fp, self, threshold)

    def threshold_tanimoto_search_arena(self, queries, threshold=0.7):
        """Find the fingerprints which are sufficiently similar to each of the query fingerprints

        For each fingerprint in the *queries* arena, find all of the
        fingerprints in this arena which are at least *threshold*
        similar. The hits are returned as a :class:`.SearchResults`,
        where the hits in each :class:`.SearchResult` is in arbitrary
        order.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResults`
        """
        self._check_at_start()
        return fps_search.threshold_tanimoto_search_arena(queries, self, threshold)

    def knearest_tanimoto_search_fp(self, query_fp, k=3, threshold=0.7):
        """Find the k-nearest fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this reader which are at least
        *threshold* similar to the query fingerprint, and of those, select
        the top *k* hits. The hits are returned as a :class:`.SearchResult`,
        sorted from highest score to lowest.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        self._check_at_start()
        return fps_search.knearest_tanimoto_search_fp(query_fp, self, k, threshold)

    def knearest_tanimoto_search_arena(self, queries, k=3, threshold=0.7):
        """Find the k-nearest fingerprints which are sufficiently similar to each of the query fingerprints

        For each fingerprint in the *queries* arena, find the
        fingerprints in this reader which are at least *threshold*
        similar to the query fingerprint, and of those, select the top
        *k* hits. The hits are returned as a :class:`.SearchResults`,
        where the hits in each :class:`.SearchResult` are sorted by
        similarity score.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResults`
        """
        self._check_at_start()
        return fps_search.knearest_tanimoto_search_arena(queries, self, k, threshold)

    def count_tversky_hits_fp(self, query_fp, threshold=0.7, alpha=1.0, beta=1.0):
        """Count the fingerprints which are sufficiently similar to the query fingerprint

        Return the number of fingerprints in the reader which are
        at least *threshold* similar to the query fingerprint *query_fp*.

        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :type alpha: float between 0.0 and 100.0, inclusive
        :type beta: float between 0.0 and 100.0, inclusive
        :returns: integer count
        """
        self._check_at_start()
        return fps_search.count_tversky_hits_fp(query_fp, self, threshold, alpha, beta)

    def threshold_tversky_search_fp(self, query_fp, threshold=0.7, alpha=1.0, beta=1.0):
        """Find the fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this reader which are at least
        *threshold* similar to the query fingerprint *query_fp*.  The
        hits are returned as a :class:`.SearchResult`, in arbitrary
        order.
        
        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :type alpha: float between 0.0 and 100.0, inclusive
        :type beta: float between 0.0 and 100.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        self._check_at_start()
        return fps_search.threshold_tversky_search_fp(query_fp, self, threshold, alpha, beta)

    def knearest_tversky_search_fp(self, query_fp, k=3, threshold=0.7, alpha=1.0, beta=1.0):
        """Find the k-nearest fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this reader which are at least
        *threshold* similar to the query fingerprint, and of those, select
        the top *k* hits. The hits are returned as a :class:`.SearchResult`,
        sorted from highest score to lowest.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :type alpha: float between 0.0 and 100.0, inclusive
        :type beta: float between 0.0 and 100.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        self._check_at_start()
        return fps_search.knearest_tversky_search_fp(query_fp, self, k, threshold, alpha, beta)

def open_output(destination, metadata, compression="", level=None, errors="strict", location=None):
    output, close = io.open_compressed_output(destination, compression, "chemfp", level=level)
    try:
        io.write_fps1_magic(output)
        initial_lineno = 1 + io.write_fps1_header(output, metadata)
        error_handler = io.get_parse_error_handler(errors)

        if location is None:
            location = io.Location.from_destination(destination)

        num_fp_bytes = -1
        if metadata is not None and metadata.num_bytes is not None:
            num_fp_bytes = metadata.num_bytes
            if not isinstance(num_fp_bytes, int):
                raise ValueError("metadata num_bytes must be an integer")
            
        writer = _fps_writer_gen(output, close, num_fp_bytes, error_handler, location, initial_lineno)
        assert next(writer) == "Ready!" # prime the generator

        return FPSWriter(output, writer, metadata, location)
    except Exception:
        if close is not None:
            close()
        raise

### The C extension which implements this function is something like:
## def write_fp_id(output, fp, id, num_fp_bytes, scratch_buffer):
##     output.write(_hexlify(fp) + ("\t" + id + "\n").encode("utf8"))
### It uses the scratch_buffer to format the fingerprint as hex and the utf8-encoded id
### before writing it to outfile.
    
def _fps_writer_gen(output, close, num_fp_bytes, error_handler, location, initial_lineno):
    recno = 0
    output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    def get_lineno():
        return initial_lineno + output_recno
    location.register(get_recno=get_recno, get_output_recno=get_output_recno, get_lineno=get_lineno)
    write_fp_id = _chemfp.write_fp_id
    
    # Adapter for Python 2.7. The underlying C code uses PyArg_ParseTuple
    # "s" to convert the id to a char *. Under Python 2.7:
    #    "Unicode objects are converted to C strings using the default encoding."
    # while under Python 3.x:
    #    "Unicode objects are converted to C strings using ``'utf-8'`` encoding."
    # My system uses a default encoding of 'ascii', so Unicode ids fail.
    # I don't really support Unicode ids under Python 2.7 for non-ascii default
    # encodings.
    try:
        write_fp_id(None, b"", u"\xff", 0, None)
    except TypeError:
        # Python 3.x allows the Unicode id, but fails in the last argument saying:
        #   TypeError: argument 5 must be read-write bytes-like object, not None
        # This is expected.
        pass
    except UnicodeEncodeError:
        # I'll convert to UTF8 first.
        def write_fp_id(output, fp, id, num_fp_bytes, scratch_buffer):
            return _chemfp.write_fp_id(output, fp, id.encode("utf8"), num_fp_bytes, scratch_buffer)
    else:
        raise AssertionError("How did this happen?")
    
    errno = 0
    try:
        yield_value = "Ready!"
        scratch_buffer = array.array("b", b"") # place to format the line before writing it.
        while 1:
            id_fp_pairs = (yield yield_value)
            try:
                for id, fp in id_fp_pairs:
                    recno += 1
                    ### Before Unicode support I formatted everything in Python. With chemfp 3.0
                    ### I switched to a C extension to format the line and write it to the file.

                    ### Here's the old code to check for illegal characters.
                    ## if "\t" in id:
                    ##     error_handler.error("Unable to write an identifier containing a tab: %r" % (id,),
                    ##                         location)
                    ##     continue
                    ## if "\n" in id:
                    ##     error_handler.error("Unable to write an identifier containing a newline: %r" % (id,),
                    ##                         location)
                    ##     continue
                    ## if not id:
                    ##     error_handler.error("Unable to write a fingerprint with an empty identifier",
                    ##                         location)
                    ##     continue
                    try:
                        ### Here's the old code to format the line.
                        ## output.write(hexlify(fp) + ("\t" + id + "\n").encode("utf8"))

                        # The new code uses a C extension.
                        errno = write_fp_id(output, fp, id, num_fp_bytes, scratch_buffer)
                        if errno == 5:
                            # Either num_fp_bytes == -1 or the scratch buffer is too small.
                            if num_fp_bytes == -1:
                                # First time through, record the expected fingerprint size.
                                num_fp_bytes = len(fp)
                            # Resize the scratch space, including some extra space
                            scratch_buffer = array.array(
                                "b", b"\0" * (num_fp_bytes*2 + 2 + len(id.encode("utf8")) + 20))
                            errno = write_fp_id(output, fp, id, num_fp_bytes, scratch_buffer)

                    # Is there a speedup? Yes. In this micro benchmark:
                    #   data = [("123", b"x"*(1024//8))] * 1000000
                    #   write_fingerprints(data)
                    # old code with only output.write(hexlify...encode("utf8")): 1.94 sec.
                    # old code also with checks for illegal characters: 2.14 sec.
                    # new code: 1.7 sec.
                    # It's very unlikely that even the old version will be a bottle-neck.

                    except UnicodeEncodeError as err:
                        error_handler.error("Cannot encode identifier %r: %s" % (id, err),
                                            location)
                        continue
                    if errno:
                        if errno == 1:
                            error_handler.error("Unable to write an identifier containing a tab: %r" % (id,),
                                                location)
                            continue
                        if errno == 2:
                            error_handler.error("Unable to write an identifier containing a newline: %r" % (id,),
                                                location)
                            continue
                        if errno == 3:
                            error_handler.error("Unable to write a fingerprint with an empty identifier",
                                                location)
                            continue
                        if errno == 4:
                            error_handler.error("Fingerprint for id %r is %d bytes long, expected %d bytes"
                                                % (id, len(fp), num_fp_bytes),
                                                location)
                            continue
                        
                        raise AssertionError("Unknown errno", errno)
                    
                    output_recno += 1
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
            else:
                yield_value = None
                
    finally:
        location.save(recno=get_recno(), output_recno=get_output_recno(), lineno=get_lineno())
        if close is not None:
            try:
                close()
            finally:
                close = None
        
class FPSWriter(FingerprintWriter):
    """Write fingerprints in FPS format.

    This is a subclass of :class:`chemfp.FingerprintWriter`.

    Instances have the following attributes:

    * metadata - a :class:`chemfp.Metadata` instance
    * format - the string 'fps'
    * closed - False when the file is open, else True
    * location - a :class:`chemfp.io.Location` instance

    An FPSWriter is its own context manager, and will close the
    output file on context exit.

    The Location instance supports the "recno", "output_recno",
    and "lineno" properties.
    """
    format = "fps"
    def __init__(self, output, writer, metadata, location=None):
        "The constructor is not part of the public API"
        self._output = output # Should this be public?
        self._writer = writer
        self.metadata = metadata
        self.location = location
        self.closed = False

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """Close the writer

        This will set self.closed to False.
        """
        if not self.closed:
            self.closed = True
            self._writer.close()


    def write_fingerprint(self, id, fp):
        """Write a single fingerprint record with the given id and fp

        :param string id: the record identifier
        :param bytes fp: the fingerprint
        """
        if self.closed:
            raise ValueError("Cannot write a fingerprint after calling close()")
        err = self._writer.send( [(id, fp)] )
        if err is not None:
            raise_tb(err[0], err[1])

    def write_fingerprints(self, id_fp_pairs):
        """Write a sequence of fingerprint records

        :param id_fp_pairs: An iterable of (id, fingerprint) pairs.
        """
        if self.closed:
            raise ValueError("Cannot write fingerprints after calling close()")
        err = self._writer.send(id_fp_pairs)
        if err is not None:
            raise_tb(err[0], err[1])



def read_header(f, location):
    #metadata = Metadata()
    num_bits = None
    num_bytes = None
    software = None
    type = None
    sources = []
    date = None
    aromaticity = None

    location.save(lineno=0)

    def error(errmsg):
        location.save(lineno=lineno+1)
        return ParseError(errmsg, location=location)

    lineno = 0
    block_reader = _read_blocks(f)
    for block in block_reader:
        # A block must be non-empty
        start = 0
        while 1:
            c = block[start:start+1]
            if c == b"":
                # End of the block; get the next one
                break
            if c != b'#':
                # End of the header. This block contains the first fingerprint line
                block = block[start:]
                if num_bits is None:
                    if num_bytes is None:
                        # We can figure this out from the fingerprint on the first line
                        err = _chemfp.fps_line_validate(-1, block)
                        if err:
                            if err == -36: # CHEMFP_MISSING_NEWLINE
                                # The block does not end with a newline.
                                # Figure out the correct line number for that line.
                                lineno += block.count(b"\n")
                            raise error(_chemfp.strerror(err))
                        i = block.index(b"\t")
                        # If you don't specify the number of bits then I'll do it for you.
                        num_bits = i * 4
                        num_bytes = i // 2
                    else:
                        num_bits = num_bytes * 8
                else:
                    if num_bytes is None:
                        num_bytes = (num_bits+7) // 8
                    else:
                        # Already validated this case when the values were set.
                        pass

                location.save(lineno=lineno)
                return Metadata(num_bits, num_bytes, type, aromaticity, software,
                                sources, date), block, block_reader

            start += 1 # Skip the '#'
            end = block.find(b"\n", start)
            if end == -1:
                raise error(_chemfp.strerror(-36)) # CHEMFP_MISSING_NEWLINE
                ## # Only happens when the last line of the file contains
                ## # no newlines. In that case, we're at the last block.
                ## line = block[start:]
                ## start = len(block)
            else:
                line = block[start:end]
                start = end+1

            # Right! We've got a line. Check if it's magic
            # This is the only line which cannot contain a '='
            if lineno == 0:
                if line == b"FPS1" or line == b"FPS1\r":
                    lineno += 1
                    continue

            if line.startswith(b"x-") or line.startswith(b"X-"):
                # Completely ignore the contents of 'experimental' lines
                continue

            if b"=" not in line:
                raise error("header line must contain an '=': %r" % (line,))
            key, value = line.split(b"=", 1)
            key = key.strip()
            value = value.strip()
            if key == b"num_bits":
                if num_bits is not None:
                    raise error("Only one num_bits header is allowed")
                try:
                    num_bits = int(value)
                    if num_bits <= 0:
                        raise ValueError
                except ValueError:
                    raise error("num_bits header must be a positive integer, not %r" % (value,))

                if num_bytes is not None:
                    n = (num_bits+7)//8
                    if n != num_bytes:
                        raise error(
                            "The num_bits header of %d requires %d bytes, which is "
                            "incompatible with the earlier num_bytes header of %d"
                            %  (num_bits, n, num_bytes))
                                        
            elif key == b"num_bytes":
                if num_bytes is not None:
                    raise error("Only one num_bytes header is allowed")
                try:
                    num_bytes = int(value)
                    if num_bytes <= 0:
                        raise ValueError
                except ValueError:
                    raise error(
                        "num_bytes header must be a positive integer, not %r" % (value,))
                if num_bits is not None:
                    n = (num_bits+7)//8
                    if n != num_bytes:
                        raise error(
                            "The num_bytes header of %d is incompatible with the "
                            "earlier num_bits header of %d, which requires %d bytes"
                            % (num_bytes, num_bits, n))
                            
            elif key == b"software":
                if software is not None:
                    raise error("Only one software header is allowed")
                try:
                    software = value.decode("utf8")
                except UnicodeDecodeError as err:
                    raise error("software header must be utf8 encoded: %r (%s)" % (value, err))
            elif key == b"type":
                if type is not None:
                    raise error("Only one type header is allowed")
                # Should I have an auto-normalization step here which
                # removes excess whitespace?
                #type = normalize_type(value)
                try:
                    type = value.decode("ascii")
                except UnicodeDecodeError as err:
                    raise error("type header must be ASCII encoded: %r (%s)" % (value, err))
            elif key == b"source":
                sources.append(value.decode("utf8", "surrogateescape"))
            elif key == b"date":
                if date is not None:
                    raise error("Only one date header is allowed")
                try:
                    date = value.decode("ascii")
                except UnicodeDecodeError as err:
                    raise error("date header must be ASCII encoded: %r (%s)" % (value, err))
            elif key == b"aromaticity":
                if aromaticity is not None:
                    raise error("Only one aromaticity header is allowed")
                try:
                    aromaticity = value.decode("ascii")
                except UnicodeDecodeError as err:
                    raise error("aromaticity header must be ASCII encoded: %r (%s)" % (value, err))
            elif key.startswith(b"x-"):
                pass
            else:
                # In the interests of forward compatibility, I ignore unknown headers.
                # XXX Should I give a warning? A log message? How does someone help
                # figure out if there's a typo?
                #print "UNKNOWN", repr(line), repr(key), repr(value)
                #warn(filename, lineno, "Unknown header %r" % (value,))
                pass
            lineno += 1

    # Reached the end of file. No fingerprint lines and nothing left to process.
    if num_bits is None:
        if num_bytes is None:
            num_bits = None
            num_bytes = None
        else:
            num_bits = num_bytes*8
    else:
        if num_bytes is None:
            num_bytes = (num_bits+7)//8
        else:
            # Already verified this case when the values were first set.
            pass

    location.save(lineno=lineno)
    return Metadata(num_bits, num_bytes, type, aromaticity, software,
                    sources, date), None, None

