# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import

import ctypes

from openeye import oechem
from openeye.oechem import OEGraphMol, OEQMol
if 1:
    from openeye.oegraphsim import *
else:
    # Suppress pyflake warnings
    from openeye.oegraphsim import (
        OEFPAtomType_Aromaticity,
        OEFPAtomType_AtomicNumber,
        OEFPAtomType_Chiral,
        OEFPAtomType_DefaultAtom,
        OEFPAtomType_DefaultCircularAtom,
        OEFPAtomType_DefaultPathAtom,
        OEFPAtomType_DefaultTreeAtom,
        OEFPAtomType_EqAromatic,
        OEFPAtomType_EqHBondAcceptor,
        OEFPAtomType_EqHBondDonor,
        OEFPAtomType_EqHalogen,
        OEFPAtomType_FormalCharge,
        OEFPAtomType_HCount,
        OEFPAtomType_HvyDegree,
        OEFPAtomType_Hybridization,
        OEFPAtomType_InRing,
        OEFPBondType_BondOrder,
        OEFPBondType_Chiral,
        OEFPBondType_DefaultBond,
        OEFPBondType_DefaultCircularBond,
        OEFPBondType_DefaultPathBond,
        OEFPBondType_DefaultTreeBond,
        OEFPBondType_InRing,
        OEFingerPrint,
        OEGetFPAtomType,
        OEGetFPBondType,
        OEGraphSimGetRelease,
        OEGraphSimGetVersion,
        OEGraphSimIsLicensed,
        OEMakeCircularFP,
        OEMakeMACCS166FP,
        OEMakePathFP,
        OEMakeTreeFP,
        )

from . import types
from . import openeye_toolkit

SOFTWARE = "OEGraphSim/%(release)s (%(version)s)" % dict(
    # There can be spaces in the beta releases
    release = OEGraphSimGetRelease().replace(" ", "_"),
    version = OEGraphSimGetVersion())
from . import SOFTWARE as CHEMFP_SOFTWARE

_has_screen = hasattr(oechem, "OESubSearchScreen")
if _has_screen:
    from openeye.oechem import (
        OESubSearchScreen, OEMakeSubSearchTargetScreen,OEMakeSubSearchQueryScreen,
        OESubSearchScreenType_Molecule, OESubSearchScreenType_SMARTS, OESubSearchScreenType_MDL
        )

############# Used when generate the FPS header


try:
    OEMakeCircularFP
except Exception:
    raise ImportError("chemfp 3.x no longer supports OEGraphSim v1")
OEGRAPHSIM_API_VERSION = "2"


_is_licensed = OEGraphSimIsLicensed()

def is_licensed():
    return _is_licensed

_available_types = []
_unavailable_types = []

def _register_v2(fingerprint_type_class, is_available=True):
    if not _is_licensed:
        _unavailable_types.append( (fingerprint_type_class.name,
                                    "OEGraphSim cannot find a valid license") )
        return
    if is_available:
        _available_types.append(fingerprint_type_class)
    else:
        name = fingerprint_type_class.name
        _unavailable_types.append(
            (name, "This version of OEGraphSim does not support the %s fingerprint" % (name,)))


##### Handle the atom and bond type flags for path fingerprints

# The atom and bond type flags can be specified on the command-line
# 
#   --atype=DefaultAtom --btype=BondOrder,InRing
#   --atype AtomicNumber,InRing  --btype DefaultBond,InRing
#
# The type fields may be separated by either a "," or a "|".
# The relevant OpenEye function (OEGetFPAtomType() and
# OEGetFPBondType()) use a "|" but that requires escaping for
# the shell, so I support "," as well.

# There's another conversion of the integer type values into a string
# representation, used when generating the canonical form of the
# generation parameters for the FPS output. That case uses "|"
# (and not ",") and omits the DefaultAtom and DefaultBond name.
# The result is easier to parse with the OpenEye API functions.


# Note: Version 1.0 of OEGraphSim uses different names than 2.0 (Grr!)
# OpenEye says these names will not change again. We'll see.

_atype_flags = [(OEGetFPAtomType(atype), atype) for atype in ( 
                    OEFPAtomType_Aromaticity,   # Arom
                    OEFPAtomType_AtomicNumber,  # AtmNum
                    OEFPAtomType_Chiral,        # Chiral
                    OEFPAtomType_EqHalogen,     # EqHalo
                    OEFPAtomType_FormalCharge,  # FCharge
                    OEFPAtomType_HvyDegree,     # HvyDeg
                    OEFPAtomType_Hybridization, # Hyb
                    OEFPAtomType_InRing,        # InRing
                    OEFPAtomType_EqAromatic,    # EqArom
                    OEFPAtomType_HCount,           # HCount
                    OEFPAtomType_EqHBondAcceptor,  # EqHBAcc
                    OEFPAtomType_EqHBondDonor,     # EqHBDon
                    )]
    
_btype_flags = [(OEGetFPBondType(btype), btype) for btype in
                (OEFPBondType_BondOrder,
                OEFPBondType_Chiral,
                 OEFPBondType_InRing)]

# I support the DefaultAtom and DefaultBond special values.
# (Note: complex bitflags go first; it simplifies the flag->description code)


_atype_default_flags = [("DefaultPathAtom", OEFPAtomType_DefaultPathAtom),
                        ("DefaultCircularAtom", OEFPAtomType_DefaultCircularAtom),
                        ("DefaultTreeAtom", OEFPAtomType_DefaultTreeAtom),
                        ("DefaultAtom", OEFPAtomType_DefaultAtom)]
_btype_default_flags = [("DefaultPathBond", OEFPBondType_DefaultPathBond),
                        ("DefaultCircularBond", OEFPBondType_DefaultCircularBond),
                        ("DefaultTreeBond", OEFPBondType_DefaultTreeBond),
                        ("DefaultBond", OEFPBondType_DefaultBond)]

_path_atype_flags = _atype_flags + _atype_default_flags + [("Default", OEFPAtomType_DefaultPathAtom)]
_path_btype_flags = _btype_flags + _btype_default_flags + [("Default", OEFPBondType_DefaultPathBond)]

_circular_atype_flags = _atype_flags + _atype_default_flags + [
    ("Default", OEFPAtomType_DefaultCircularAtom)]
_circular_btype_flags = _btype_flags + _btype_default_flags + [
    ("Default", OEFPBondType_DefaultCircularBond)]

_tree_atype_flags = _atype_flags + _atype_default_flags + [("Default", OEFPAtomType_DefaultTreeAtom)]
_tree_btype_flags = _btype_flags + _btype_default_flags + [("Default", OEFPBondType_DefaultTreeBond)]

_path_atypes = dict(_path_atype_flags)
_path_btypes = dict(_path_btype_flags)
_circular_atypes = dict(_circular_atype_flags)
_circular_btypes = dict(_circular_btype_flags)
_tree_atypes = dict(_tree_atype_flags)
_tree_btypes = dict(_tree_btype_flags)


## Go from a "," or "|" separated text field to an integer value
# Removes extra whitespace, but none should be present.

def _get_type_value(a_or_b, table, description):
    value = 0
    # Allow both "|" and "," as separators
    # (XXX OEGraphSim 2.0.0 only allows "|")
    description = description.replace("|", ",")
    for word in description.split(","):
        word = word.strip()
        try:
            value |= table[word]
        except KeyError:
            if not word:
                raise ValueError("Missing %s flag" % (a_or_b,))
            raise ValueError("Unknown %s type %r (expecting one of: %s)" % (
                a_or_b, word, ", ".join(sorted(table))))
    return value

def path_atom_description_to_value(description):
    """path_atom_description_to_value(description) -> integer

    Convert an atom description like FormalCharge,EqHalogen
    or FormalCharge|EqHalogen into its atom type value.

    This is similar to OEGetFPAtomType except both "|" and "," are
    allowed seperators and "AtomDefault" is an allowed term.
    """
    return _get_type_value("path atom", _path_atypes, description)

def path_bond_description_to_value(description):
    """bond_description_to_value(description) -> integer

    Convert an bond description like BondOrder,Chiral
    or BondOrder|Chiral into its bond type value.

    This is similar to OEGetFPBondType except both "|" and "," are
    allowed seperators and "BondDefault" is an allowed term.
    """
    return _get_type_value("path bond", _path_btypes, description)

def circular_atom_description_to_value(description):
    return _get_type_value("circular atom", _circular_atypes, description)
def circular_bond_description_to_value(description):
    return _get_type_value("circular bond", _circular_btypes, description)
def tree_atom_description_to_value(description):
    return _get_type_value("tree atom", _tree_atypes, description)
def tree_bond_description_to_value(description):
    return _get_type_value("tree bond", _tree_btypes, description)

## Go from an integer value into a canonical description

# I could use OEGetFPAtomType() and OEGetFPBondType() but I wanted
# something which has a fixed sort order even for future releases,
# which isn't part of those functions.

def _get_type_description(a_or_b, flags, value):
    words = []
    for (word, flag) in flags:
        if flag & value == flag:
            # After over 12 years of full-time use of Python,
            # I finally have a chance to use the xor operator.
            value = value ^ flag
            words.append(word)
    if value != 0:
        raise AssertionError("Unsupported %s value %d" % (a_or_b, value))
    return "|".join(words)


def path_atom_value_to_description(value):
    """atom_value_to_description(value) -> string

    Convert from an atom type string into its text description,
    separated by "|"s. The result are compatible with
    OEGetFPAtomType and are in canonical order.
    """
    return _get_type_description("path atom", _path_atype_flags, value)

def path_bond_value_to_description(value):
    """bond_value_to_description(value) -> string

    Convert from a bond type string into its text description,
    separated by "|"s. The result are compatible with
    OEGetFPBontType and are in canonical order.
    """
    return _get_type_description("path bond", _path_btype_flags, value)

def circular_atom_value_to_description(value):
    return _get_type_description("circular atom", _circular_atype_flags, value)
def circular_bond_value_to_description(value):
    return _get_type_description("circular bond", _circular_btype_flags, value)
def tree_atom_value_to_description(value):
    return _get_type_description("tree atom", _tree_atype_flags, value)
def tree_bond_value_to_description(value):
    return _get_type_description("tree bond", _tree_btype_flags, value)

    
##### Create a function which generate fingerprints

# I use functions which return functions because it was a nice way to
# hide the differences between the two fingerprinters. I also found
# that I can save a bit of time by not creating a new fingerprint each
# time. The measured speedup is about 2% for MACCS166 and 6% for path
# fingerprints.

# Just like the OEGraphMol, these fingerprints must not be reused or
# stored. They are mutated EVERY TIME. They are NOT thread-safe.
# If you need to use these in multiple threads, then make multiple
# fingerprinters.

# Believe it or not, reusing the preallocated fingerprint measurably
# helps the performance.

def get_path_fingerprinter(numbits, minbonds, maxbonds, atype, btype):
    # Extra level of error checking since I expect people will think
    # of this as part of the public API.
    if not (16 <= numbits <= 65536):
        raise ValueError("numbits must be between 16 and 65536 (inclusive)")
    if not (0 <= minbonds):
        raise ValueError("minbonds must be 0 or greater")
    if not (minbonds <= maxbonds):
        raise ValueError("maxbonds must not be smaller than minbonds")

    # XXX validate the atype and type values?
    # It's a simple mask against the | of all possible value, then test for 0.
    # However, I'm not sure what to report as the error message.
    
    fp = OEFingerPrint()
    fp.SetSize(numbits)
    data_location = int(fp.GetData())
    num_bytes = (numbits+7)//8
    empty_fp = b"\0" * num_bytes

    def path_fingerprinter(mol):
        if not OEMakePathFP(fp, mol, numbits, minbonds, maxbonds, atype, btype):
            return empty_fp
        return ctypes.string_at(data_location, num_bytes)

    return path_fingerprinter

def get_maccs_fingerprinter():
    fp = OEFingerPrint()
    # Call SetSize() now to force space allocation, so I only need one GetData()
    fp.SetSize(166)
    data_location = int(fp.GetData())
    num_bytes = (166+7)//8
    empty_fp = b"\0" * num_bytes
    
    def maccs_fingerprinter(mol):
        if not OEMakeMACCS166FP(fp, mol):
            return empty_fp
        return ctypes.string_at(data_location, num_bytes)

    return maccs_fingerprinter

def get_circular_fingerprinter(numbits, minradius, maxradius, atype, btype):
    # Extra level of error checking since I expect people will think
    # of this as part of the public API.
    if not (16 <= numbits <= 65536):
        raise ValueError("numbits must be between 16 and 65536 (inclusive)")
    if not (0 <= minradius):
        raise ValueError("minradius must be 0 or greater")
    if not (minradius <= maxradius):
        raise ValueError("maxradius must not be smaller than minradius")

    fp = OEFingerPrint()
    fp.SetSize(numbits)
    data_location = int(fp.GetData())
    num_bytes = (numbits+7)//8
    empty_fp = b"\0" * num_bytes

    def circular_fingerprinter(mol):
        if not OEMakeCircularFP(fp, mol, numbits, minradius, maxradius, atype, btype):
            return empty_fp
        return ctypes.string_at(data_location, num_bytes)

    return circular_fingerprinter


def get_tree_fingerprinter(numbits, minbonds, maxbonds, atype, btype):
    # Extra level of error checking since I expect people will think
    # of this as part of the public API.
    if not (16 <= numbits <= 65536):
        raise ValueError("numbits must be between 16 and 65536 (inclusive)")
    if not (0 <= minbonds):
        raise ValueError("minbonds must be 0 or greater")
    if not (minbonds <= maxbonds):
        raise ValueError("maxbonds must not be smaller than minbonds")

    fp = OEFingerPrint()
    fp.SetSize(numbits)
    data_location = int(fp.GetData())
    num_bytes = (numbits+7)//8
    empty_fp = b"\0" * num_bytes

    def tree_fingerprinter(mol):
        if not OEMakeTreeFP(fp, mol, numbits, minbonds, maxbonds, atype, btype):
            return empty_fp
        return ctypes.string_at(data_location, num_bytes)


    return tree_fingerprinter


def get_molecule_screener():
    screen = OESubSearchScreen()
    numbits = 896
    screen.SetSize(numbits)
    data_location = int(screen.GetData())
    num_bytes = (numbits+7)//8
    empty_fp = b"\0" * num_bytes
    
    def oechem_molecule_screener(mol):
        if isinstance(mol, OEQMol):
            if not OEMakeSubSearchQueryScreen(screen, mol, OESubSearchScreenType_Molecule):
                return empty_fp
        else:
            if not OEMakeSubSearchTargetScreen(screen, mol, OESubSearchScreenType_Molecule):
                return empty_fp
        return ctypes.string_at(data_location, num_bytes)
    
    return oechem_molecule_screener

def get_smarts_screener():
    screen = OESubSearchScreen()
    numbits = 896
    screen.SetSize(numbits)
    data_location = int(screen.GetData())
    num_bytes = (numbits+7)//8
    empty_fp = b"\0" * num_bytes

    def oechem_smarts_screener(mol):
        if isinstance(mol, OEQMol):
            if not OEMakeSubSearchQueryScreen(screen, mol, OESubSearchScreenType_SMARTS):
                return empty_fp
        else:
            if not OEMakeSubSearchTargetScreen(screen, mol, OESubSearchScreenType_SMARTS):
                return empty_fp
        return ctypes.string_at(data_location, num_bytes)

    return oechem_smarts_screener

def get_mdl_screener():
    screen = OESubSearchScreen()
    numbits = 896
    screen.SetSize(numbits)
    data_location = int(screen.GetData())
    num_bytes = (numbits+7)//8
    empty_fp = b"\0" * num_bytes

    def oechem_mdl_screener(mol):
        if isinstance(mol, OEQMol):
            if not OEMakeSubSearchQueryScreen(screen, mol, OESubSearchScreenType_MDL):
                return empty_fp
        else:
            if not OEMakeSubSearchTargetScreen(screen, mol, OESubSearchScreenType_MDL):
                return empty_fp
        return ctypes.string_at(data_location, num_bytes)

    return oechem_mdl_screener

### A note on fingerprints and ctypes.string_at

# The FPS format and OEFingerPrint.GetData() values used identical bit
# and byte order. Bytes are in little-endian order and bits are in
# big-endian order. That means I can use GetData() to get the
# underlying C storage area, use ctypes to turn that into a Python
# string, which I then hex encode.

# The other option is to use OEFingerPrint.ToHexString(). But that's
# pure little endian, so I would need a transposition to make the bits
# be what I want them to be. OEChem's hex strings also end with a flag
# which says how many extra bits to trim, which I don't need since I
# handle it a different way.

# Here's some info about the bit order, which I tested by setting a
# few bits though the API then seeing what changed in the hex string
# and in the underlying GetData() field.

# The bit pattern
#   01234567 89ABCDEF  pure little endian
#   10011100 01000011
#   
#         93 2C    using ToHexString()  (pure little endian)
#       0x39 c2    using hex(ord(GetData())) (litle endian byte, big endian bit)
#
#   76543210 FEDCBA98
#   00111001 11000010   little endian byte, big endian bit



#####  Configure the fingerprints

# Don't mark this as thread-safe because chemfp.openeye_toolkit defines
# these types but uses a delayed loader.
class BaseOpenEyeFingerprintType(types.FingerprintType):
    software = SOFTWARE + " " + CHEMFP_SOFTWARE
    toolkit = openeye_toolkit

class VariableSizeFingerprint(types.ThreadsafeFingerprinterMixin, BaseOpenEyeFingerprintType):
    def __init__(self, fingerprint_kwargs):
        super(VariableSizeFingerprint, self).__init__(fingerprint_kwargs)
        self.num_bits = fingerprint_kwargs["numbits"]

class FixedSizeFingerprint(types.ThreadsafeFingerprinterMixin, BaseOpenEyeFingerprintType):
    pass

class OpenEyeMACCSFingerprintType_v2(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """OEGraphSim implementation of the 166 MACCS keys, version 2

    See https://docs.eyesopen.com/toolkits/cpp/graphsimtk/fingerprint.html#maccs .

    The OpenEye-MACCS166/2 :class:`.FingerprintType` has no parameters.

    This corresponds to GraphSim version '2.0.0'.
    """
    name = "OpenEye-MACCS166/2"
    num_bits = 166
    _make_fingerprinter = staticmethod(get_maccs_fingerprinter)


if _is_licensed:
    _PROBE_MOL = OEGraphMol()
    _PROBE_FP = OEFingerPrint()
    _PROBE_MOL.NewAtom(2)
    OEMakeMACCS166FP(_PROBE_FP, _PROBE_MOL)
    _maccs_version = {
        "2.0.0": "2",
        "2.2.0": "3",
        }.get(_PROBE_FP.GetFPTypeBase().GetFPVersionString(), "unknown")
else:
    _maccs_version = "unknown"

_register_v2(OpenEyeMACCSFingerprintType_v2, _maccs_version=="2")

class OpenEyeMACCSFingerprintType_v3(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """OEGraphSim implementation of the 166 MACCS keys, version 3

    See https://docs.eyesopen.com/toolkits/cpp/graphsimtk/fingerprint.html#maccs .

    The OpenEye-MACCS166/3 :class:`.FingerprintType` has no parameters.

    This corresponds to GraphSim version '2.2.0', with fixes for bits 91 and 92.
    """
    name = "OpenEye-MACCS166/3"
    num_bits = 166
    _make_fingerprinter = staticmethod(get_maccs_fingerprinter)

_register_v2(OpenEyeMACCSFingerprintType_v3, _maccs_version=="3")

def _correct_numbits(s):
    try:
        if not s.isdigit():
            raise ValueError
        i = int(s)
        if not (16 <= i <= 65536):
            raise ValueError
    except ValueError:
        raise ValueError("must be between 16 and 65536 bits")
    return i


path_arguments = types.FingerprintArguments()
_add = path_arguments.add_argument
_add("numbits", decoder=_correct_numbits, metavar="INT", default=4096,
     help="number of bits in the path fingerprint")

_add("minbonds", decoder=types.nonnegative_int, metavar="INT", default=0,
     help="minimum number of bonds in the path fingerprint")

_add("maxbonds", decoder=types.nonnegative_int, metavar="INT", default=5,
     help="maximum number of bonds in the path fingerprint")

_add("atype", decoder=path_atom_description_to_value,
     encoder=path_atom_value_to_description,
    help="atom type as a '|' separated list of terms", default=OEFPAtomType_DefaultAtom)

_add("btype", decoder=path_bond_description_to_value,
     encoder=path_bond_value_to_description,
    help="bond type as a '|' separated list of terms", default=OEFPBondType_DefaultBond)

class OpenEyePathFingerprintType_v2(VariableSizeFingerprint):
    """OEGraphSim fingerprint based on path-based enumeration, version 2
    
    See https://docs.eyesopen.com/toolkits/cpp/graphsimtk/fingerprint.html#section-fingerprint-path

    The OpenEye-Path/2 :class:`.FingerprintType` parameters are:

    * numbits - the number of bits in the fingerprint (default: 4096)
    * minbonds - the minimum number of bonds (default: 0)
    * maxbonds - the maximum number of bonds (default: 5)
    * atype - the atom type (default: "Default")
    * btype - the bond type (default: "Default")

    The atype is either 0 or a '|' separated string containing one
    or more of the following: Aromaticity, AtomicNumber, Chiral,
    EqHBondAcceptor, EqHBondDonor, EqHalogen, FormalCharge, HCount,
    HvyDegree, Hybridization, InRing, EqAromatic,
    
    The btype is either 0 or a '|' separated string containing one
    or more of the following: BondOrder, Chiral, InRing.
    """
    name = "OpenEye-Path/2"
    _format_string = ("numbits={numbits} minbonds={minbonds} "
                      "maxbonds={maxbonds} atype={atype} btype={btype}")
    _arguments = path_arguments
    _make_fingerprinter = staticmethod(get_path_fingerprinter)

_register_v2(OpenEyePathFingerprintType_v2)


circular_arguments = types.FingerprintArguments()
_add = circular_arguments.add_argument
_add("numbits", decoder=_correct_numbits, metavar="INT", default=4096,
     help="number of bits in the circular fingerprint")
_add("minradius", decoder=types.nonnegative_int, metavar="INT", default=0,
     help="minimum radius for the circular fingerprint")
_add("maxradius", decoder=types.nonnegative_int, metavar="INT", default=5,
     help="maximum radius for the circular fingerprint")
_add("atype", decoder=circular_atom_description_to_value,
     encoder=circular_atom_value_to_description,
     help="atom type", default=OEFPAtomType_DefaultCircularAtom)
_add("btype", decoder=circular_bond_description_to_value,
     encoder=circular_bond_value_to_description,
     help="bond type", default=OEFPBondType_DefaultCircularBond)

class OpenEyeCircularFingerprintType_v2(VariableSizeFingerprint):
    """OEGraphSim fingerprint based on circular fingerprints around heavy atoms, version 2
    
    See https://docs.eyesopen.com/toolkits/cpp/graphsimtk/fingerprint.html#section-fingerprint-circular

    The OpenEye-Circular/2 :class:`.FingerprintType` parameters are:

    * numbits - the number of bits in the fingerprint (default: 4096)
    * minradius - the minimum radius (default: 0)
    * maxradius - the maximum radius (default: 5)
    * atype - the atom type (default: "Default")
    * btype - the bond type (default: "Default")

    The atype is either 0 or a '|' separated string containing one
    or more of the following: Aromaticity, AtomicNumber, Chiral,
    EqHBondAcceptor, EqHBondDonor, EqHalogen, FormalCharge, HCount,
    HvyDegree, Hybridization, InRing, EqAromatic,
    
    The btype is either 0 or a '|' separated string containing one
    or more of the following: BondOrder, Chiral, InRing.
    """
    name = "OpenEye-Circular/2"
    _arguments = circular_arguments
    _format_string = ("numbits={numbits} minradius={minradius} "
                      "maxradius={maxradius} atype={atype} btype={btype}")
    _make_fingerprinter = staticmethod(get_circular_fingerprinter)

_register_v2(OpenEyeCircularFingerprintType_v2)

tree_arguments = types.FingerprintArguments()
_add = tree_arguments.add_argument
_add("numbits", decoder=_correct_numbits, metavar="INT", default=4096,
     help="number of bits in the tree fingerprint")
_add("minbonds", decoder=types.nonnegative_int, metavar="INT", default=0,
     help="minimum number of bonds in the tree fingerprint")
_add("maxbonds", decoder=types.nonnegative_int, metavar="INT", default=4,
     help="maximum number of bonds in the tree fingerprint")
_add("atype", decoder=tree_atom_description_to_value,
     encoder=tree_atom_value_to_description,
     help="atom type", default=OEFPAtomType_DefaultTreeAtom)
_add("btype", decoder=tree_bond_description_to_value,
     encoder=tree_bond_value_to_description,
     help="bond type", default=OEFPBondType_DefaultTreeBond)

class OpenEyeTreeFingerprintType_v2(VariableSizeFingerprint):
    """OEGraphSim fingerprint based on tree fingerprints, version 2
    
    See https://docs.eyesopen.com/toolkits/cpp/graphsimtk/fingerprint.html#section-fingerprint-tree

    The OpenEye-Tree/2 :class:`.FingerprintType` parameters are:

    * numbits - the number of bits in the fingerprint (default: 4096)
    * minbonds - minimum number of bonds in the tree
    * maxbonds - maximum number of bonds in the tree
    * atype - the atom type (default: "Default")
    * btype - the bond type (default: "Default")

    The atype is either 0 or a '|' separated string containing one
    or more of the following: Aromaticity, AtomicNumber, Chiral,
    EqHBondAcceptor, EqHBondDonor, EqHalogen, FormalCharge, HCount,
    HvyDegree, Hybridization, InRing, EqAromatic,
    
    The btype is either 0 or a '|' separated string containing one
    or more of the following: BondOrder, Chiral, InRing.
    """
    name = "OpenEye-Tree/2"
    _arguments = tree_arguments
    _format_string = ("numbits={numbits} minbonds={minbonds} "
                      "maxbonds={maxbonds} atype={atype} btype={btype}")
    _make_fingerprinter = staticmethod(get_tree_fingerprinter)


_register_v2(OpenEyeTreeFingerprintType_v2)

class OpenEyeMoleculeScreenFingerprintType_v1(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """OEChem molecule screen using OESubSearchScreenType::Molecule
    
    See http://https://docs.eyesopen.com/toolkits/cpp/oechemtk/OEChemClasses/OESubSearchScreen.html 
    This OpenEyeMoleculeScreenFingerprintType_v1 :class:`.FingerprintType`
    takes no parameters. Calling the fingerprinter with a QMol returns the
    query screen, calling with an OEMol returns a target screen.
    """
    name = "OpenEye-MoleculeScreen/1"
    num_bits = 896
    _make_fingerprinter = staticmethod(get_molecule_screener)

_register_v2(OpenEyeMoleculeScreenFingerprintType_v1, _has_screen)

class OpenEyeSMARTSScreenFingerprintType_v1(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """OEChem SMARTS screen using OESubSearchScreenType::SMARTS
    
    See http://https://docs.eyesopen.com/toolkits/cpp/oechemtk/OEChemClasses/OESubSearchScreen.html 
    This OpenEyeSMARTSScreenFingerprintType_v1 :class:`.FingerprintType`
    takes no parameters. Calling the fingerprinter with a QMol returns the
    query screen, calling with an OEMol returns a target screen.
    """
    name = "OpenEye-SMARTSScreen/1"
    num_bits = 896
    _make_fingerprinter = staticmethod(get_smarts_screener)

_register_v2(OpenEyeSMARTSScreenFingerprintType_v1, _has_screen)

class OpenEyeMDLScreenFingerprintType_v1(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """OEChem MDL screen using OESubSearchScreenType::MDL
    
    See http://https://docs.eyesopen.com/toolkits/cpp/oechemtk/OEChemClasses/OESubSearchScreen.html 
    This OpenEyeMDLScreenFingerprintType_v1 :class:`.FingerprintType`
    takes no parameters. Calling the fingerprinter with a QMol returns the
    query screen, calling with an OEMol returns a target screen.
    """
    name = "OpenEye-MDLScreen/1"
    num_bits = 896
    _make_fingerprinter = staticmethod(get_mdl_screener)

_register_v2(OpenEyeMDLScreenFingerprintType_v1, _has_screen)


def resolve_chemfp_fingerprint_types(registry):
    for name, reason in _unavailable_types:
        registry.register_unavailable(name, reason)
        
    for fingerprint_type in _available_types:
        registry.register_fingerprint_type(fingerprint_type)

