# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""text_toolkit - record-oriented structure toolkit I/O library

The text_toolkit implements the chemfp toolkit API but where the
"molecules" are simple TextRecord instances which store the
records as text strings. It does not use a back-end chemistry toolkit,
and it cannot convert between different chemistry representations.

The TextRecord is a base class. The actual records depend on the
format, and will be one of:

* :class:`.SDFRecord`
* :class:`.SmiRecord`
* :class:`.CanRecord`
* :class:`.UsmRecord`
* :class:`.SmiStringRecord`
* :class:`.CanStringRecord`
* :class:`.UsmStringRecord`

The text toolkit will let you "convert" between the different SMILES
formats, but it doesn't actually change the SMILES string. The SMILES
records have the attributes ``id``, ``record`` and ``smiles``.

The toolkit also knows a bit about the SD format. The SDF records have
the attributes ``id``, ``id_bytes`` and ``record``, and there are
methods to get SD tag values and add a tag to the end of the tag data
block.

The text_toolkit also supports a few SDF-specific I/O functions to read
SDF records directly as a string instead of wrapped in a TextRecord.

The record types also have the attributes ``encoding`` and
``encoding_errors`` which affect how the record bytes are parsed.

"""

from __future__ import absolute_import

from . import _text_toolkit as _tk

__all__ = [
    "is_licensed",
    "get_formats",
    "get_input_formats",
    "get_output_formats",
    "get_format",
    "get_input_format",
    "get_output_format",
    "get_input_format_from_source",
    "get_output_format_from_destination",
    "read_molecules",
    "read_molecules_from_string",
    "read_ids_and_molecules",
    "read_ids_and_molecules_from_string",
    "make_id_and_molecule_parser",
    "parse_molecule",
    "parse_id_and_molecule",
    "create_string",
    "create_bytes",
    "open_molecule_writer",
    "open_molecule_writer_to_string",
    "open_molecule_writer_to_bytes",
    "copy_molecule",
    "add_tag",
    "get_tag",
    "get_tag_pairs",
    "get_id",
    "set_id",

    "read_sdf_records",
    "read_sdf_ids_and_records",
    "read_sdf_ids_and_values",
    "read_sdf_records_from_string",
    "read_sdf_ids_and_records_from_string",
    "read_sdf_ids_and_values_from_string",
    "get_sdf_tag",
    "add_sdf_tag",
    "get_sdf_tag_pairs",
    "get_sdf_id",
    "set_sdf_id",
]

name = "text"

from . import SOFTWARE as software  # eg, chemfp/3.3

def is_licensed():
    """Return True - chemfp's text toolkit is always licensed

    :returns: True
    """
    return True

_format_registry = _tk.format_registry
_toolkit = _tk.toolkit

# format API

def get_formats(include_unavailable=False):
    """Get the list of structure formats that chemfp's text toolkit supports

    This version of chemfp will always support the structure formats
    available to chemfp so 'include_unavailable' does not affect anything.
    (It may affect other toolkits.)

    :param include_unavailable: include unavailable formats?
    :value include_unavailable: True or False
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_formats(include_unavailable)

def get_input_formats():
    """Get the list of supported chemfp text toolkit input formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_formats()

def get_output_formats():
    """Get the list of supported chemfp text toolkit output formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_formats()

def get_format(format_name):
    """Get the named format, or raise a ValueError

    This will raise a ValueError for unknown format names.

    :param format_name: the format name
    :value format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_format(format_name)

def get_input_format(format_name):
    """Get the named input format, or raise a ValueError

    This will raise a ValueError for unknown format names
    or if that format is not an input format.

    :param format_name: the format name
    :value format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_input_format(format_name)

def get_output_format(format_name):
    """Get the named format, or raise a ValueError

    This will raise a ValueError for unknown format names
    or if that format is not an output format.
    
    :param format_name: the format name
    :value format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_output_format(format_name)

def get_input_format_from_source(source=None, format=None):
    """Get the most appropriate format given the available source and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *source* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param source: The structure data source.
    :type source: A filename (as a string), a file object, or None to read from stdin
    :param format: Format information, if known.
    :type format: A Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_input_format_from_source(source, format)

def get_output_format_from_destination(destination=None, format=None):
    """Get the most appropriate format given the available destination and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *destination* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param destination: The structure data source.
    :type destination: A filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: A Format(-like) object, string, or None
    :returns: A :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_output_format_from_destination(destination, format)

# SDF-specific API

def read_sdf_records(source=None, reader_args=None, compression=None,
                     errors="strict", location=None, block_size=327680):
    """Return an iterator that reads each record from an SD file as a string.

    Iterate through the records in *source*, which must be in SD format.
    If *compression* is None or "auto" then auto-detect the compression type
    based on *source*, and default to uncompressed when it can't be determined.
    Use "gz" when the input is gzip compressed, and "none" or "" if uncompressed.

    The *reader_args* parameter is currently unused. It exists for future compatability.

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.

    The *block_size* parameter is the number of bytes to read from the SD file.
    The current implementation reads a block, iterates through the records in the
    block, then prepends any remaining text to the start of the next block. You
    shouldn't need to change this parameter, but if you do, please let me know.

    Note: to prevent accidental memory consumption if the input is in the wrong
    format, a complete record must be found within the first 327680 bytes or
    5*block_size bytes, whichever is larger.

    The parser has only a basic understanding of the SD format. It knows how
    to handle the counts line, the SKP property, and even tag data with
    the value '$$$$'. It is not a full validator and it does not know chemistry.

    See :func:`.read_sdf_ids_and_records` if you want (id, record)
    pairs, and :func:`.read_sdf_ids_and_values` if you want
    (id, tag data) pairs. See
    :func:`.read_sdf_ids_and_records_from_string` to read from
    a string instead of a file or file-like object.

    :param source: the SDF source
    :type source: a filename, file object, or None to read from stdin
    :param reader_args: currently ignored
    :type reader_args: currently ignored
    :param compression: the data content compression method
    :type compression: one of "auto", "none", "", or "gz"
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :func:`chemfp.base_toolkit.RecordReader` iterating over the records as a string
    """
    return _tk.read_sdf_records(source, reader_args, compression, errors, location, block_size)
    

def read_sdf_ids_and_records(source=None, id_tag=None, reader_args=None, compression=None,
                            errors="strict", location=None,
                            encoding="utf8", encoding_errors="strict", block_size=327680):
    """Return an iterator that reads the (id, record string) pairs from an SD file

    See :func:`.read_sdf_records` for most parameter details. That function
    iterates over the records, while this one iterates over the (id, record)
    pairs. By default the id comes from the title line. Use *id_tag* to get
    the record id from the given SD tag instead.

    See :func:`.read_sdf_ids_and_values` if you want to read an identifier
    and tag value, or two tag values, instead of returning the full record.

    :param source: the SDF source
    :type source: a filename, file object, or None to read from stdin
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: currently ignored
    :type reader_args: currently ignored
    :param compression: the data content compression method
    :type compression: one of "auto", "none", "", or "gz"
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndRecordReader` iterating (id, record string) pairs
    """
    return _tk.read_sdf_ids_and_records(source, id_tag, reader_args, compression, errors, location,
                                        encoding, encoding_errors, block_size)

def read_sdf_ids_and_values(source=None, id_tag=None, value_tag=None,
                            reader_args=None, compression=None, errors="strict", location=None,
                            encoding="utf8", encoding_errors="strict", block_size=327680):
    """Return an iterator that reads the (id, tag value string) pairs from an SD file

    See :func:`.read_sdf_records` for most parameter details. That function
    iterates over the records, while this one iterates over the (id, tag value)
    pairs.

    By default this uses the title line for both the id and tag value
    strings.  Use *id_tag* and *value_tag*, respectively, to use a
    given tag value instead. If a tag doesn't exist then None will
    be used.

    :param source: the SDF source
    :type source: a filename, file object, or None to read from stdin
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param value_tag: SD tag containing the value
    :type value_tag: string, or None to use the record title
    :param reader_args: currently ignored
    :type reader_args: currently ignored
    :param compression: the data content compression method
    :type compression: one of "auto", "none", "", or "gz"
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndRecordReader` iterating (id, value string) pairs
    """
    return _tk.read_sdf_ids_and_values(source, id_tag, value_tag, reader_args, compression,
                                       errors, location, encoding, encoding_errors, block_size)


def read_sdf_records_from_string(content, reader_args=None, compression=None,
                                 errors="strict", location=None, block_size=327680):
    """Return an iterator that reads each record from a string containing SD records

    See :func:`.read_sdf_records_from_string` for the parameter details. The
    main difference is that this function reads from *content*, which is a
    string containing 0 or more SDF records.

    If content is a (Unicode) string then it must only contain ASCII characters,
    the records will be returned as strings, and the compression option is not
    supported. If content is a byte string then the records will be returned as
    byte strings, and compression is supported.
    
    See :func:`.read_sdf_ids_and_records_from_string` to read (id, record)
    pairs and :func:`.read_sdf_ids_and_values_from_string` to read (id, tag value)
    pairs.

    :param content: a string containing zero or more SD records
    :type content: string or bytes
    :param reader_args: currently ignored
    :type reader_args: currently ignored
    :param compression: the data content compression method
    :type compression: one of "auto", "none", "", or "gz"
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.RecordReader` iterating over each record as a string
    """

    return _tk.read_sdf_records_from_string(content, reader_args, compression, errors, location, block_size)

def read_sdf_ids_and_records_from_string(content=None, id_tag=None, reader_args=None,
                                         compression=None, errors="strict", location=None,
                                         encoding="utf8", encoding_errors="strict",
                                         block_size=327680):
    """Return an iterator that reads the (id, record) pairs from a string containing SD records

    This function reads the records from *content*, which is a string containing
    0 or more SDF records. It iterates over the (id, record) pairs. By default
    the id comes from the first line of the SD record. Use *id_tag* to use
    a given tag value instead. See :func:`.read_sdf_records` for details about
    the other parameters.

    If content is a (Unicode) string then it must only contain ASCII characters,
    the records will be returned as strings, the compression option is not
    supported, and the encoding and encoding_errors parameters are ignored.
    
    If content is a byte string then the records will be returned as
    byte strings, compression is supported, and the encoding and encoding_errors
    parameters are used to parse the id.
    
    :param content: a string containing zero or more SD records
    :type content: string or bytes
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: currently ignored
    :type reader_args: currently ignored
    :param compression: the data content compression method
    :type compression: one of "auto", "none", "", or "gz"
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndRecordReader` iterating over the (id, record string) pairs
    """
    return _tk.read_sdf_ids_and_records_from_string(content, id_tag, reader_args, compression,
                                                    errors, location, encoding, encoding_errors, block_size)

def read_sdf_ids_and_values_from_string(content=None, id_tag=None, value_tag=None,
                                        compression=None, reader_args=None, errors="strict",
                                        location=None, encoding="utf8", encoding_errors="strict",
                                        block_size=327680):
    """Return an iterator that reads the (id, value) pairs from a string containing SD records

    This function reads the records from *content*, which is a string
    containing 0 or more SDF records. It iterates over the (id, value)
    pairs, which by default both contain the title line. Use *id_tag*
    and *value_tag*, respectively, to use a given tag value instead.
    If a tag doesn't exist then None will be used.

    If content is a (Unicode) string then it must only contain ASCII characters, the
    compression option is not supported, and the encoding and encoding_errors
    parameters are ignored.
    
    If content is a byte string then the records will be returned as
    byte strings, compression is supported, and the encoding and encoding_errors
    parameters are used to parse the id and value.

    See :func:`.read_sdf_records` for details about the other parameters.

    :param content: a string containing zero or more SD records
    :type content: string or bytes
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param value_tag: SD tag containing the value
    :type value_tag: string, or None to use the record title
    :param reader_args: currently ignored
    :type reader_args: currently ignored
    :param compression: the data content compression method
    :type compression: one of "auto", "none", "", or "gz"
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndRecordReader` iterating over the (id, value) pairs
    """
    return _tk.read_sdf_ids_and_values_from_string(content, id_tag, value_tag, compression,
                                                   reader_args, errors, location,
                                                   encoding, encoding_errors, block_size)

# read/write API

def read_molecules(source=None, format=None, id_tag=None, reader_args=None,
                   errors="strict", location=None,
                   encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads TextRecord instances from a structure file

    Iterate through the *format* structure records in *source*. If *format*
    is None then auto-detect the format based on the *source*. For SD files,
    use *id_tag* to get the record id from the given SD tag instead of the
    title line. (read_molecules() will ignore the *id_tag*. It exists to
    make it easier to switch between reader functions.)

    Only the SMILES formats use the *reader_args* dictionary. The supported
    parameters are:
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * has_header - True or False

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.

    See :func:`.read_ids_and_molecules` if you want (id, :class:`.TextRecord`)
    pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader parameters passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param encoding: the byte encoding
    :type encoding: string (typically 'utf8' or 'latin1')
    :param encoding_errors: how to handle decoding failure
    :type encoding_errors: string (typically 'strict', 'ignore', or 'replace')
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating :class:`.TextRecord` molecules
    """
    return _toolkit.read_molecules(source, format, id_tag, reader_args, errors, location, encoding, encoding_errors)

def read_molecules_from_string(content, format, id_tag=None, reader_args=None,
                               errors="strict", location=None):
    """Return an iterator that reads TextRecord instances from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`.read_molecules` for details about the other parameters. See
    :func:`.read_ids_and_molecules_from_string` if you want to read (id, :class:`.TextRecord`)
    pairs instead of just molecules.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param encoding: the byte encoding
    :type encoding: string (typically 'utf8' or 'latin1')
    :param encoding_errors: how to handle decoding failure
    :type encoding_errors: string (typically 'strict', 'ignore', or 'replace')
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating :class:`.TextRecord` molecules
    """
    return _toolkit.read_molecules_from_string(content, format, id_tag, reader_args,
                                               errors, location)

def read_ids_and_molecules(source=None, format=None, id_tag=None, reader_args=None,
                           errors="strict", location=None,
                           encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads (id, TextRecord) pairs from a structure file

    See :func:`chemfp.text_toolkit.read_molecules` for full parameter details.
    The major difference is that this returns an iterator of (id, :class:`.TextRecord`)
    pairs instead of just the molecules.
    
    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param encoding: the byte encoding
    :type encoding: string (typically 'utf8' or 'latin1')
    :param encoding_errors: how to handle decoding failure
    :type encoding_errors: string (typically 'strict', 'ignore', or 'replace')
    :returns: a :class:`chemfp.text_toolkit.IdAndMoleculeReader` iterating (id, :class:`.TextRecord`) pairs
    """
    return _toolkit.read_ids_and_molecules(source, format, id_tag, reader_args, errors, location,
                                           encoding, encoding_errors)

def read_ids_and_molecules_from_string(content, format, id_tag=None, reader_args=None,
                                       errors="strict", location=None):
    """Return an iterator that reads (id, TextRecord) pairs from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.rdkit_toolkit.read_molecules` for details about the other
    parameters. See :func:`chemfp.rdkit_toolkit.read_molecules_from_string` if you
    just want to read the :class:`.TextRecord` molecules instead of (id, TextRecord) pairs.
    
    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param encoding: the byte encoding
    :type encoding: string (typically 'utf8' or 'latin1')
    :param encoding_errors: how to handle decoding failure
    :type encoding_errors: string (typically 'strict', 'ignore', or 'replace')
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, :class:`.TextRecord`) pairs
    """
    return _toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                       errors, location)

def make_id_and_molecule_parser(format, id_tag=None, reader_args=None, errors="strict"):
    """Create a specialized function which takes a record and returns an (id, TextRecord) pair

    The returned function is optimized for reading many records from individual
    strings because it only does parameter validation once. However, I haven't
    really noticed much of a performance difference between this and
    :func:`chemfp.text_toolkit.parse_id_and_molecule` so I suggest you use
    that function directly instead of making a specialized function.
    (Let me know if making a specialized function is useful.)

    See :func:`chemfp.text_toolkit.read_molecules` for details about the
    other parameters. The specific :class:`.TextRecord` subclass returned
    depends on the format.

    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a function of the form ``parser(record string) -> (id, text_record)``
    """
    return _toolkit.make_id_and_molecule_parser(format, id_tag, reader_args, errors)

def parse_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from the *content* string and return a TextRecord.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.text_toolkit.read_molecules`
    for details about the other parameters. See :func:`chemfp.text_toolkit.parse_id_and_molecule`
    if you want the (id, :class:`.TextRecord`) pair instead of just the text record.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param encoding: the byte encoding
    :type encoding: string (typically 'utf8' or 'latin1')
    :param encoding_errors: how to handle decoding failure
    :type encoding_errors: string (typically 'strict', 'ignore', or 'replace')
    :returns: a :class:`.TextRecord`
    """
    return _toolkit.parse_molecule(content, format, id_tag, reader_args, errors)

def parse_id_and_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from *content* and return the (id, TextRecord) pair.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.rdkit_toolkit.read_molecules`
    for details about the other parameters.

    See :func:`chemfp.rdkit_toolkit.read_molecules` for details about the
    other parameters. See :func:`chemfp.rdkit_toolkit.parse_molecule`
    if just want the :class:`.TextRecord` and not the the (id, TextRecord) pair.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param encoding: the byte encoding
    :type encoding: string (typically 'utf8' or 'latin1')
    :param encoding_errors: how to handle decoding failure
    :type encoding_errors: string (typically 'strict', 'ignore', or 'replace')
    :returns: an (id, :class:`.TextRecord` molecule) pair
    """
    return _toolkit.parse_id_and_molecule(content, format, id_tag, reader_args, errors)

def create_string(mol, format, id=None, writer_args=None, errors="strict"):
    """Convert a TextRecord into a structure record in the given format as a Unicode string

    If *id* is not None then use it instead of the molecule's own id.

    :param mol: the molecule to use for the output
    :type mol: a :class:`.TextRecord`
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a Unicode string
    """
    return _toolkit.create_string(mol, format, id, writer_args, errors)

def create_bytes(mol, format, id=None, writer_args=None, errors="strict", level=None):
    """Convert a TextRecord into a structure record in the given format as a byte string

    If *id* is not None then use it instead of the molecule's own id.

    :param mol: the molecule to use for the output
    :type mol: a :class:`.TextRecord`
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a byte string
    """
    return _toolkit.create_bytes(mol, format, id, writer_args, errors, level)


def open_molecule_writer(destination=None, format=None, writer_args=None,
                         errors="strict", location=None,
                         encoding="utf8", encoding_errors="strict", level=None):
    """Return a MoleculeWriter which can write TextRecord instances to a destination.

    A :class:`chemfp.base_toolkit.MoleculeWriter` has the methods ``write_molecule``,
    ``write_molecules``, and ``write_ids_and_molecules``, which are ways to write
    an :class:`.TextRecord`, an TextRecord iterator, or an (id, TextRecord) pair
    iterator to a file.

    TextRecords are written to *destination*. The output format can be a
    string like "sdf.gz" or "smi", a :class:`chemfp.base_toolkit.Format`,
    or Format-like object with "name" and "compression" attributes, or None
    to auto-detect based on the *destination*. If auto-detection is not
    possible, the output will be written as uncompressed SMILES.

    That said, the text toolkit doesn't know how to convert between SMILES
    and SDF formats, and will raise an exception if you try.

    The *writer_args* is only used for the "smi", "can", and "usm" output
    formats. The only supported parameter is::
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.
    
    :param destination: the structure destination
    :type destination: a filename, file object, or None to write to stdout
    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param encoding: the byte encoding
    :type encoding: string (typically 'utf8' or 'latin1')
    :param encoding_errors: how to handle decoding failure
    :type encoding_errors: string (typically 'strict', 'ignore', or 'replace')
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting :class:`.TextRecord` instances
    """
    return _toolkit.open_molecule_writer(destination, format, writer_args, errors, location,
                                         encoding, encoding_errors, level)
                                         

def open_molecule_writer_to_string(format, writer_args=None, errors="strict", location=None):
    """Return a MoleculeStringWriter which can write TextRecord instances to a string.

    See :func:`chemfp.text_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a Unicode string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting :class:`.TextRecord` instances
    """
    return _toolkit.open_molecule_writer_to_string(format, writer_args, errors, location)

def open_molecule_writer_to_bytes(format, writer_args=None, errors="strict", location=None, level=None):
    """Return a MoleculeStringWriter which can write TextRecord instances to a string.

    See :func:`chemfp.text_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a byte string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting :class:`.TextRecord` instances
    """
    return _toolkit.open_molecule_writer_to_bytes(format, writer_args, errors, location, level)


# molecule manipulation API

def copy_molecule(mol):
    """Return a new TextRecord which is a copy of the given TextRecord
    
    :param mol: the text record
    :type mol: a :class:`.TextRecord`
    :returns: a new :class:`.TextRecord`
    """
    return mol.copy()

def add_tag(mol, tag, value):
    """Add an SD tag value to the TextRecord

    If the *mol* is in "sdf" format then this will modify
    ``mol.record`` to append the new *tag* and *value* to the
    end of the tag block. The other tags will not be modified,
    including tags with the same tag name.

    :param mol: the text record
    :type mol: a :class:`.TextRecord`
    :param string tag: the SD tag name
    :param string value: the text for the tag
    :returns: None
    """
    mol.add_tag(tag, value)

def get_tag(mol, tag):
    """Get the named SD tag value, or None if it doesn't exist

    If the *mol* is in "sdf" format then this will return
    the corresponding tag value from ``mol.record``, or None
    if the tag does not exist.

    If the record is in any other format then it will return None.

    :param mol: the molecule
    :type mol: a :class:`.TextRecord`
    :param tag: the SD tag name
    :type tag: string
    :returns: a string, or None
    """
    return mol.get_tag(tag)


def get_tag_pairs(mol):
    """Get a list of all SD tag (name, value) pairs for the TextRecord

    If the *mol* is in "sdf" format then this will return
    the list of (tag, value) pairs in ``mol.record``, where the
    *tag* and *value* are strings.

    If the record is in any other format then it will return
    an empty list.

    :param mol: the molecule
    :type mol: a :class:`.TextRecord`
    :returns: a list of (tag name, tag value) pairs
    """
    return mol.get_tag_pairs()

def get_id(mol):
    """Get the molecule's id from the TextRecord's id field

    This is toolkit-portable way to get ``mol.id``.
    
    :param mol: the molecule
    :type mol: a TextRecord
    :returns: a string
    """
    return mol.id

def set_id(mol, id):
    """Set the TextRecord's id to the new id

    This is the toolkit-portable way to write ``mol.id = id``.

    Note: this does not modify ``mol.record``. Use :func:`chemfp.text_toolkit.create_string`
    or similar text_toolkit functions to get the record text with a new identifier.

    :param mol: the molecule
    :type mol: a :class:`.TextRecord`
    :param id: the new id
    :type id: string
    :returns: None
    """
    # I don't set the SD record title here. It's set when I generate the output.
    # (Is that reasonable?)
    if "\n" in id:
        raise ValueError("Id must not contain a newline")
    if "\t" in id:
        raise ValueError("Id must not contain a tab character")
    mol.id = id

# SDF record manipulation API
def get_sdf_tag(sdf_record, tag):
    """Return the value for a named tag in an SDF record string

    Get the value for the tag named *tag* from the string *sdf_record*
    containing an SD record.

    :param string sdf_record: an SD record
    :param string tag: a tag name
    :returns: the corresponding tag value as a string, or None
    """
    return _tk.get_sdf_record_tag(sdf_record, tag)

def add_sdf_tag(sdf_record, tag, value):
    """Add an SD tag value to an SD record string

    This will append the new *tag* and *value* to the
    end of the tag data block in the *sdf_record* string.
    
    :param string sdf_record: an SD record
    :param string tag: a tag name
    :param string value: the new tag value
    :returns: a new SD record string with the new tag and value
    """
    return _tk.add_sdf_record_tag(sdf_record, tag, value)

def get_sdf_tag_pairs(sdf_record):
    """Return the (tag, value) entries in the SDF record string

    Parse the *sdf_record* and return the tag data as a list of (tag,
    value) pairs. The type of the returned strings will be the same
    as the type of the input sdf_record string.

    :param string sdf_record: an SDF record
    :returns: a list of (tag, value) pairs
    """
    return _tk.get_sdf_tag_pairs(sdf_record)

def get_sdf_id(sdf_record):
    """Return the id for the SDF record string

    The id is the first line of the *sdf_record*. A future version
    of this function may support an *id_tag* parameter. Let me
    know if that would be useful.

    The returned id string will have the same type as the input
    sdf_record.

    :param string sdf_record: an SD record
    :returns: the first line of the SD record
    """
    if isinstance(sdf_record, type(u"")):
        return sdf_record.partition(u"\n")[0]
    else:
        return sdf_record.partition(b"\n")[0]

def set_sdf_id(sdf_record, id):
    """Set the id of the SDF record string to a new value

    Set the first line of *sdf_record* to the new *id*, which must
    not contain a newline.
    
    The sdf_record and the id must have the same string type.
    
    :param string sdf_record: an SDF record
    :param string id: the new id
    """
    unicode_type = type(u"")
    if isinstance(sdf_record, unicode_type):
        if not isinstance(id, unicode_type):
            raise ValueError("Cannot mix string and bytes for sdf_record and id")
        left, mid, right = sdf_record.partition(u"\n")
        if u"\n" in id:
            raise ValueError("Cannot set id because it contains a newline")
        if mid != u"\n" or not right:
            raise ValueError("Cannot set id because the record does not appear to be in SDF format")
        return id + mid + right
    else:
        if isinstance(id, unicode_type):
            raise ValueError("Cannot mix bytes and string for sdf_record and id")
        left, mid, right = sdf_record.partition(b"\n")
        if b"\n" in id:
            raise ValueError("Cannot set id because it contains a newline")
        if mid != b"\n" or not right:
            raise ValueError("Cannot set id because the record does not appear to be in SDF format")
        return id + mid + right
